/* 
  Copyright (c) 2013 - 2014, Oracle and/or its affiliates. All rights reserved.
  
  This program is free software; you can redistribute it and/or modify it under 
  the terms of the GNU General Public License as published by the Free Software 
  Foundation; version 2 of the License.
  
  This program is distributed in the hope that it will be useful, but WITHOUT ANY 
  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS 
  FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
  
  You should have received a copy of the GNU General Public License along with this 
  program; if not, write to the Free Software Foundation, Inc., 51 Franklin St, Fifth 
  Floor, Boston, MA 02110-1301  USA
*/

#include "JCallback.h"

JCallback::JCallback()
{
  jstack= new stack_jelem[MAX_JSON_DEPTH];
  num_elem = 0;
};


JCallback::~JCallback()
{
  delete[] jstack;
};


bool JCallback::put_elem(int key_start, int key_len, int val_start, int val_len, int type_flag)
{
  stack_jelem sj;
  sj.key_start=    key_start;
  sj.key_len=      key_len;
  sj.val_start=    val_start;
  sj.val_len=      val_len;
  sj.type_flag=    type_flag;
  
  jstack[num_elem++]= sj;
  
  return true;
};


bool JCallback::check_order(int start_ind)
{
  unsigned int max= jstack[start_ind].val_start;
  for (unsigned int i= 1; i + start_ind < num_elem; i++)
  {
    if (max <= jstack[start_ind + i].val_start)
    {
      jdump(JSON_NOTE, JSON_CALLBACK_ERROR, "wrong order of elements, start_ind: %d, max: %d, cur: %d", start_ind, max, jstack[start_ind + i].val_start);
      return false;
    }
    max= jstack[start_ind + i].val_start;
  }
  
  return true;
};


bool JCallback_valid::callback_func(int key_start, int key_len, int val_start, int val_len, int pos, bool is_array_elem, int type_flag)
{
  if (0 == num_elem)
    put_elem(key_start, key_len, val_start, val_len);
  
  return true;
};

bool JCallback_test_parser::callback_func(int key_start, int key_len, int val_start, int val_len, int pos, bool is_array_elem, int type_flag)
{
  jdump(JSON_NOTE, JSON_CALLBACK_ERROR, "Pos: %d", pos);
  put_elem(key_start, key_len, val_start, val_len);
  return true;
};


char* JCallback_test_parser::get_parse_tree(char* doc)
{
  char* result= new char[MAX_JSON_DEPTH * 1024];
  int pos= 0;
  
  for (unsigned int i= 0; i < num_elem; i++)
  {
    stack_jelem js= jstack[i];
    memcpy(&result[pos], &doc[js.key_start], js.key_len);
    result[pos + js.key_len]= ' ';
    result[pos + js.key_len + 1]= '=';
    result[pos + js.key_len + 2]= '>';
    result[pos + js.key_len + 3]= ' ';
    pos= pos + js.key_len + 4;
    memcpy(&result[pos], &doc[js.val_start], js.val_len);
    result[pos + js.val_len]= ';';
    pos= pos + js.val_len + 1;
  }
  result[pos]= '\0';
  
  return result;
};


JCallback_contains_key::JCallback_contains_key(char* doc, JKey** keys, int klen)
{
  this->doc=              doc;
  this->keys=             keys;
  this->klen=             klen;
  this->key_exists=       true;
  this->ind=              klen;
  this->last_proceed_ind= klen + 1;
  this->arr_ind=          0;
  this->arr_inds=         new int[MAX_JSON_DEPTH];
  this->proceed_further=  false;
  this->proceeding_array= false;
  this->skip_stack_elem=  0;
  this->prev_pos=         -1;
  this->jstack_arrays=    new stack_jelem*[klen];
  for (int i= 0; i < klen; i++)
  {
    this->jstack_arrays[i]= NULL;
  }
  for (int i= 0; i < MAX_JSON_DEPTH; i++)
  {
    this->arr_inds[i]= 0;
  }
};


JCallback_contains_key::~JCallback_contains_key()
{
  for (int i= 0; i < klen; i++)
    if (NULL != this->jstack_arrays[i])
      delete[] this->jstack_arrays[i];
  delete[] this->jstack_arrays;
  delete[] this->arr_inds;
};


bool JCallback_contains_key::put_array_elem(int cur_pos, int arr_ind, int key_start, int key_len, int val_start, int val_len, int type_flag)
{
  if (0 == arr_inds[cur_pos])
    this->jstack_arrays[cur_pos]= new stack_jelem[MAX_JSON_DEPTH];
  
  stack_jelem sj;
  sj.key_start=    key_start;
  sj.key_len=      key_len;
  sj.val_start=    val_start;
  sj.val_len=      val_len;
  sj.type_flag=    type_flag;
  
  this->jstack_arrays[cur_pos][arr_inds[cur_pos]]= sj;

  return true;
  
};


bool JCallback_contains_key::do_callback(int key_start, int key_len, int val_start, int val_len, int pos, bool is_array_elem, int type_flag)
{
  if (0 == ind || 0 == pos)
  {
    return false;
  }
  /* We just proceed array element and found no match */
  else if (proceeding_array && ind < pos 
    && (pos - 2 < klen && (keys[pos - 2])->as_longlong() >= arr_inds[pos - 1] ) 
    && last_proceed_ind >= pos
    )
  {
    this->skip_stack_elem+= (klen - ind);
    ind= klen;
    last_proceed_ind= -1;
  }
  if (this->proceeding_array && !is_array_elem && pos < klen && -1 > (keys[pos])->as_longlong())
  {
    if (0 <= arr_inds[pos] + 1 + (keys[pos])->as_longlong())
    {
      put_elem(jstack_arrays[pos][arr_inds[pos] + 1 + (keys[pos])->as_longlong()].val_start
        , jstack_arrays[pos][arr_inds[pos] + 1 + (keys[pos])->as_longlong()].key_len
        , jstack_arrays[pos][arr_inds[pos] + 1 + (keys[pos])->as_longlong()].val_start
        , jstack_arrays[pos][arr_inds[pos] + 1 + (keys[pos])->as_longlong()].val_len
        , jstack_arrays[pos][arr_inds[pos] + 1 + (keys[pos])->as_longlong()].type_flag
        );
      key_exists= true;
      //arr_inds[pos]= 0;
      last_proceed_ind= ind;
      ind--;
    }
  }
  if (this->prev_pos > pos)
    arr_inds[prev_pos]= 0;
  this->prev_pos= pos;
  if (ind == pos)
  {
    if (!is_array_elem)
    {
      this->proceeding_array= false;
      bool test_passed= true;
      char* key= (keys[pos - 1])->as_string();
      if (NULL == key)
      {
        key_exists= false;
        return false;
      }
      else if (strlen(key) != key_len - 2)
      {
        key_exists= false;
        if (last_proceed_ind >= pos)
        {
          this->skip_stack_elem+= (klen - ind);
          ind= klen;
          last_proceed_ind= -1;
        }
        return true;
      }
      for (int i=0; i < strlen(key); i++)
      {
        if (doc[key_start + 1 + i] != key[i])
        {
          test_passed= false;
          break;
        }
      }
      key_exists= test_passed;
      if (test_passed)
      {
        put_elem(key_start, key_len, val_start, val_len, type_flag);
        last_proceed_ind= ind;
        ind--;
      }
    }
    else
    {
      this->proceeding_array= true;
      if ((keys[pos - 1])->as_longlong() == arr_inds[pos])
      {
        put_elem(val_start, key_len, val_start, val_len, type_flag);
        last_proceed_ind= ind;
        ind --;
        key_exists= true;//(ind == klen);
        arr_inds[pos]= 0;
        this->proceeding_array= false;
      }
      else if ((keys[pos - 1])->as_longlong() > arr_inds[pos])
        arr_inds[pos]++;
      else if (-1 > (keys[pos - 1])->as_longlong())
      {
        put_array_elem(pos-1, arr_inds[pos - 1], val_start, key_len, val_start, val_len, type_flag);
        arr_inds[pos - 1]++;
      }
      else
        key_exists= false;
    }
    proceed_further= true;
    return true;
  }
  else
  {
    if (is_array_elem)
      arr_inds[pos]++;
    proceed_further= false;
    return true;
  }
};


bool JCallback_contains_key::callback_func(int key_start, int key_len, int val_start, int val_len, int pos, bool is_array_elem, int type_flag)
{
  return do_callback(key_start, key_len, val_start, val_len, pos, is_array_elem, type_flag);
};


bool JCallback_contains_key::contains()
{
  return this->key_exists && ind == 0 && check_order(this->skip_stack_elem);
};


JCallback_extract::JCallback_extract(char* doc, JKey** keys, int klen)
  :JCallback_contains_key(doc, keys, klen)
{
  this->value=      NULL;
};


bool JCallback_extract::callback_func(int key_start, int key_len, int val_start, int val_len, int pos, bool is_array_elem, int type_flag)
{
  return do_callback(key_start, key_len, val_start, val_len, pos, is_array_elem, type_flag);
};


char* JCallback_extract::extract()
{
  if (this->key_exists && 0 == this->ind && check_order(this->skip_stack_elem))
  {
    int val_len= jstack[this->skip_stack_elem].val_len;
    int val_start= jstack[this->skip_stack_elem].val_start;
    if (jstring == jstack[this->skip_stack_elem].type_flag)
    {
      val_len= val_len - 2;
      val_start+= 1;
    }
    this->value= new char[val_len + 1];
    for (unsigned int i= 0; i < val_len; i++)
      this->value[i]= doc[val_start + i];
    this->value[val_len]= '\0';
    
    return this->value;
  }
  else
  {
    if (this->value)
      delete[] this->value;
    return NULL;
  }
};


JCallback_remove::JCallback_remove(char* doc, JKey** keys, int klen)
  :JCallback_contains_key(doc, keys, klen)
{
  this->key_start=  0;
  this->key_len=    0;
  this->val_start=  0;
  this->val_len=    0;
};


bool JCallback_remove::callback_func(int key_start, int key_len, int val_start, int val_len, int pos, bool is_array_elem, int type_flag)
{
  if (do_callback(key_start, key_len, val_start, val_len, pos, is_array_elem, type_flag))
  {
    if (proceed_further && key_exists && pos == klen)
    {
      proceed_further= false;
      this->key_start= key_start;
      this->val_start= val_start;
      this->val_len=   val_len;
      this->key_len=   key_len;
      
      if (is_array_elem)
      {
        this->key_start= val_start;
      }
    }
    return true;
  }
  else
  {
    return false;
  }
};


void JCallback_remove::prepare_doc(bool change_key, int val_offset)
{
  int i= 0;
  bool last_element= false;
  
  while(i < strlen(doc) - this->val_start - this->val_len)
  {
    if (',' == doc[this->val_start + this->val_len + i])
    {
      this->val_len+= i + val_offset;
      break;
    }
    else
    {
      bool break_loop= false;
      switch (doc[this->val_start + this->val_len + i])
      {
      case ' ':
      case '\n':
      case '\t':
      case '\r':
        break;
      case ']':
      case '}':
        last_element= true;
      default:
        break_loop= true;
        break;
      }
      if (break_loop)
        break;
    }
    i++;
  }
  if (last_element/*  && 0 == comma_offset */)
  {
    i= 1;
    while (i < key_start)
    {
      if (',' == doc[key_start - i])
      {
        if (change_key)
          this->key_start= key_start - i;
        break;
      }
      else
      {
        bool break_loop= false;
      switch (doc[key_start - i])
      {
      case ' ':
      case '\n':
      case '\t':
      case '\r':
        break;
      default:
        break_loop= true;
        break;
      }
      if (break_loop)
        break;
      }
      i++;
    }
  }
};


char* JCallback_remove::get_doc()
{
  char* retval;
  if (this->key_exists && 0 == this->ind && check_order(this->skip_stack_elem))
  {
    this->val_len= jstack[this->skip_stack_elem].val_len;
    this->val_start= jstack[this->skip_stack_elem].val_start;
    this->key_len= jstack[this->skip_stack_elem].key_len;
    this->key_start= jstack[this->skip_stack_elem].key_start;
    
    prepare_doc(true, 1);
    
    retval= new char[strlen(doc) - this->val_len - this->val_start + this->key_start + 1];
    memcpy(&retval[0], &doc[0], this->key_start);
    memcpy(&retval[this->key_start], &doc[this->val_start + this->val_len], strlen(doc) - this->val_len - this->val_start);
    retval[strlen(doc) - this->val_len - this->val_start + this->key_start]= '\0';
  }
  else
  {
    retval= new char[strlen(doc) + 1];
    memcpy(retval, doc, strlen(doc) + 1);
  }
  return retval;
};


JCallback_replace::JCallback_replace(char* doc, JKey** keys, int klen, char* new_value):JCallback_remove(doc, keys, klen)
{
  this->new_value= new_value;
};


char* JCallback_replace::get_doc()
{
  char* retval;
  int key_offset= 0;
  
  if (this->key_exists && 0 == this->ind && check_order(this->skip_stack_elem))
  {
    this->val_len= jstack[this->skip_stack_elem].val_len;
    this->val_start= jstack[this->skip_stack_elem].val_start;
    this->key_len= jstack[this->skip_stack_elem].key_len;
    this->key_start= jstack[this->skip_stack_elem].key_start;
    
    prepare_doc();
    
    if (0 < this->key_len)
      key_offset= 1;
    
    retval= new char[strlen(doc) - this->val_len - this->val_start + this->key_start + this->key_len + 1 + strlen(new_value) + key_offset];
    memcpy(&retval[0], &doc[0], this->key_start + this->key_len + key_offset);
    memcpy(&retval[this->key_start + this->key_len + key_offset], 
           &new_value[0], strlen(new_value));
    memcpy(&retval[this->key_start + this->key_len + key_offset + strlen(new_value)], 
           &doc[this->val_start + this->val_len], strlen(doc) - this->val_len - this->val_start);
    retval[strlen(doc) - this->val_len - this->val_start + this->key_start + this->key_len + strlen(new_value) + key_offset]= '\0';
  }
  else
  {
    retval= new char[strlen(doc) + 1];
    memcpy(retval, doc, strlen(doc) + 1);
  }
  return retval;
};


JCallback_append::JCallback_append(char* doc, JKey** keys, int klen, char* new_value)
  :JCallback_replace(doc, keys, klen, new_value)
{
  this->new_value=      new_value;
  this->last_found=     false;
  this->maximum_passed= false;
  this->in_array=       false;
  this->decrement=      1;
  this->json_set=       false;
  this->val_len=        strlen(doc);
  this->elem_found=     false;
  //this->prev_pos=       -1;
  this->last_is_empty=  false;
};


bool JCallback_append::callback_func(int key_start, int key_len, int val_start,
  int val_len, int pos, bool is_array_elem, int type_flag)
{
  if (ind - 1 == pos && ind == klen && maximum_passed)
    ind-= this->decrement;
  if (pos == klen)
    maximum_passed= true;
  if ((jobject == type_flag || jarray == type_flag) && -1 == prev_pos && pos == klen - 1)
  {
    ind--;
    last_is_empty= true;
  }
  if (0 == ind || 0 == pos)
  {
    return false;
  }
  else if (proceeding_array && ind < pos /* && klen != pos */
    && (pos - 2 < klen && (keys[pos - 2])->as_longlong() >= arr_inds[pos - 1] ) 
    && last_proceed_ind >= pos
    )
  {
    if (ind + this->decrement < pos)
      this->skip_stack_elem+= (klen - ind);
    ind= klen;
    last_found= false;
    last_proceed_ind= -1;
  }
  if (this->proceeding_array && !is_array_elem && pos < klen && -1 > (keys[pos])->as_longlong())
  {
    if (0 <= arr_inds[pos] + 1 + (keys[pos])->as_longlong())
    {
      put_elem(jstack_arrays[pos][arr_inds[pos] + 1 + (keys[pos])->as_longlong()].val_start
        , jstack_arrays[pos][arr_inds[pos] + 1 + (keys[pos])->as_longlong()].key_len
        , jstack_arrays[pos][arr_inds[pos] + 1 + (keys[pos])->as_longlong()].val_start
        , jstack_arrays[pos][arr_inds[pos] + 1 + (keys[pos])->as_longlong()].val_len
        , jstack_arrays[pos][arr_inds[pos] + 1 + (keys[pos])->as_longlong()].type_flag
        );
      this->proceeding_array= false;
      if (pos + 1 == klen)
      {
        if (this->json_set)
          this->decrement= 0;
        last_found= true;
      }
      ind --;
      key_exists= (this->json_set && last_found) || !last_found;
      last_proceed_ind= ind;
    }
    else
    {
      key_exists= false;
      return false;
    }
  }
  if (this->prev_pos > pos)
    arr_inds[prev_pos]= 0;
  this->prev_pos= pos;
  if (ind == pos)
  {
    bool put_el= false;
    if (!is_array_elem)
    {
      this->proceeding_array= false;
      bool test_passed= true;
      char* key= (keys[pos - 1])->as_string();
      if (NULL == key)
      {
        key_exists= false;
        return false;
      }
      else if (strlen(key) != key_len - 2)
      {
        test_passed= false;
        if (last_proceed_ind >= pos)
        {
          this->skip_stack_elem+= (klen - ind);
          ind= klen;
          last_proceed_ind= -1;
        }
      }
      else
      {
        for (int i=0; i < strlen(key); i++)
        {
          if (doc[key_start + 1 + i] != key[i])
          {
            test_passed= false;
            break;
          }
        }
      }
      key_exists= (this->json_set && test_passed)
                  || ((test_passed && pos != klen)
                  || (!test_passed && pos == klen));
      if (test_passed)
      {
        put_el= key_exists;
        //put_elem(key_start, key_len, val_start, val_len);
        if (pos == klen)
        {
          if (this->json_set)
            this->decrement= 0;
          last_found= true;
        }
        last_proceed_ind= ind;
        ind--;
      }
    }
    else
    {
      this->proceeding_array= true;
      if ((keys[pos - 1])->as_longlong() == arr_inds[pos])
      {
        put_el= true;
        this->proceeding_array= false;
        //put_elem(val_start, key_len, val_start, val_len);
        if (pos == klen)
        {
          if (this->json_set)
            this->decrement= 0;
          last_found= true;
        }
        last_proceed_ind= ind;
        ind --;
        key_exists= (this->json_set && last_found) || !last_found;
        arr_inds[pos] = 0;
      }
      else if ((keys[pos - 1])->as_longlong() > arr_inds[pos])
        arr_inds[pos]++;
      else if (-1 > (keys[pos - 1])->as_longlong())
      {
        jdump(JSON_NOTE, JSON_CALLBACK_ERROR, "saving array elem, arr_ind: %d", arr_inds[pos - 1]);
        put_array_elem(pos-1, arr_inds[pos - 1], val_start, key_len, val_start, val_len, type_flag);
        arr_inds[pos - 1]++;
      }
      else
        key_exists= false;
    }
    if (put_el/* key_exists && pos == klen - this->decrement */)
    {
      elem_found= true;
      
      if (is_array_elem)
        put_elem(val_start, key_len, val_start, val_len);
      else
        put_elem(key_start, key_len, val_start, val_len);
    }
    return true;
  }
  else
  {
    if (is_array_elem)
      arr_inds[pos]++;
    return true;
  }
};


int JCallback_append::prepare_value()
{
  int retval= -1;
  int vlen=   this->val_len - 1;
  
  while (0 < vlen)
  {
    if ('}' == this->doc[this->val_start + vlen]
      || ']' == this->doc[this->val_start + vlen])
    {
      if (']' == this->doc[this->val_start + vlen])
        this->in_array= true;
      retval= this->val_start + vlen;
      break;
    }
    else
    {
      switch(this->doc[this->val_start + vlen])
      {
      case ' ':
      case '\n':
      case '\r':
      case '\t':
        break;
      default:
        return -1;
      }
    }
    vlen--;
  }
  
  return retval;
}


char* JCallback_append::get_doc()
{
  char* retval;
  int key_offset= 0;
  int bracket= -1;
  
  if (this->key_exists && 0 == this->ind && !this->last_found && check_order(this->skip_stack_elem))
  {
    if (elem_found)
    {
      this->val_len= jstack[this->skip_stack_elem].val_len;
      this->val_start= jstack[this->skip_stack_elem].val_start;
      this->key_len= jstack[this->skip_stack_elem].key_len;
      this->key_start= jstack[this->skip_stack_elem].key_start;
    }
    
    prepare_doc();
    
    bracket= prepare_value();
    
    if (0 > bracket)
    {
      //TODO: return error - parent is scalar
      return NULL;
    }
    
    if (in_array)
    {
      int comma_offset= (last_is_empty ? 0 : 2);
      retval= new char[strlen(doc) + 1 + strlen(new_value) + comma_offset];
      memcpy(&retval[0], &doc[0], bracket);
      if (!last_is_empty)
        memcpy(&retval[bracket], ", ", 2);
      else
        jdump(JSON_NOTE, JSON_CALLBACK_ERROR, "last is empty");
      memcpy(&retval[bracket + comma_offset], &new_value[0], strlen(new_value));
      memcpy(&retval[bracket + comma_offset + strlen(new_value)], &doc[bracket], strlen(doc) - bracket);
      retval[strlen(doc) + strlen(new_value) + comma_offset]= '\0';
    }
    else
    {
      char* key= keys[klen - 1]->as_string();
      if (NULL != key)
      {
        int comma_offset= (last_is_empty ? 0 : 2);
        retval= new char[strlen(doc) + 1 + strlen(new_value) + comma_offset + strlen(key) + 4];
        memcpy(&retval[0], &doc[0], bracket);
        if (last_is_empty)
          memcpy(&retval[bracket], "\"", 1);
        else
          memcpy(&retval[bracket], ", \"", 3);
        memcpy(&retval[bracket + comma_offset + 1], key, strlen(key));
        memcpy(&retval[bracket + comma_offset + 1 + strlen(key)], "\": ", 3);
        memcpy(&retval[bracket + comma_offset + 4 + strlen(key)], &new_value[0], strlen(new_value));
        memcpy(&retval[bracket + comma_offset + 4 + strlen(key) + strlen(new_value)], &doc[bracket], strlen(doc) - bracket);
        retval[strlen(doc) + strlen(new_value) + comma_offset + 4 + strlen(key)]= '\0';
      }
      else
      {
        jdump(JSON_ERROR, JSON_CALLBACK_ERROR, "Numeric key instead of string inside object!");
        retval= NULL;
      }
    }
  }
  else
  {
    retval= new char[strlen(doc) + 1];
    memcpy(retval, doc, strlen(doc) + 1);
  }
  return retval;
};



JCallback_set::JCallback_set(char* doc, JKey** keys, int klen, char* new_value):JCallback_append(doc, keys, klen, new_value)
{
  this->json_set= true;
};


char* JCallback_set::get_doc()
{
  if (!this->last_found)
    return JCallback_append::get_doc();
  else
    return JCallback_replace::get_doc();
};


JCallback_search::JCallback_search(char* doc, char*value, int search_pos)
{
  this->value=      value;
  this->doc=        doc;
  this->path=       new JPath*[MAX_JSON_DEPTH];
  this->path_len=   0;
  this->exists=     false;
  this->cur_pos=    -1;
  this->arr_ind=    -1;
  this->search_pos= search_pos;
};


bool JCallback_search::callback_func(int key_start, int key_len, int val_start, int val_len, int pos, bool is_array_elem, int type_flag)
{
  if (is_array_elem)
    arr_ind++;
  
  int j= 0;   /* Position of a char in the value to find */
  bool treat___as_real= false;
  bool treat_percent_as_real= treat_percent_as_real;
  bool any_char= false;
  int percent_pos= -1;
  
  if (!this->exists)
  {
    for (int i= 0; i < val_len; i++)
    {
      if ('\\' == value[j])
      {
        j++;
        if ('_' == value[j])
          treat___as_real= true;
        if ('%' == value[j])
          treat_percent_as_real= true;
      }
      
      if ('%' == value[j] && !treat_percent_as_real)
      {
        any_char= true;
        percent_pos= j;
        j++;
      }
      
      if (doc[val_start + i] == value[j])
      {
        if (any_char)
          any_char= false;
      }
      else if ('_' != value[j] || treat___as_real)
      {
        if (0 > percent_pos)
          return true;
        if (!any_char)
        {
          j= percent_pos + 1;
          any_char= true;
        }
        if (doc[val_start + i] == value[j])
          j++;
      }
      
      if (!any_char)
        j++;
      treat___as_real= false;
      treat_percent_as_real= false;
    }
    
    if (j == strlen(this->value) && 0 == (--this->search_pos))
      exists= true;
    else
      return true;
  }
  
  //exists
  if (-1 == cur_pos || pos < cur_pos)
  {
    if (!is_array_elem && 2 <= key_len)
    {
      path[path_len]= new JPath(new char[key_len - 2], key_len - 2);
      
      memcpy(&(path[path_len]->key[0]), &doc[key_start + 1], key_len - 2);
      path_len++;
      arr_ind= -1;
    }
    else if (is_array_elem)
    {
      char* key= new char[33];
      sprintf(key, "%d", this->arr_ind);
      path[path_len++]= new JPath(key, strlen(key), true);
    }
    cur_pos= pos;
  }
  
  return true;
};


char* JCallback_search::get_path(bool path_format)
{
  if (path_format)
  {
    if (this->exists && 0 < this->path_len)
    {
      char* result= new char[strlen(doc)]; //change to actual path len
      int res_len= 1;
      result[0]= '$';
      for (int i= path_len - 1; i >= 0; i--)
      {
        if (!path[i]->in_array)
        {
          result[res_len]= '.';
          char* temp_res= escape(path[i]->key, path[i]->key_len, (const char*) ".[]");
          memcpy(&result[res_len + 1], temp_res, strlen(temp_res));
          res_len+= (strlen(temp_res) + 1);
          if (temp_res)
            delete[] temp_res;
        }
        else
        {
          result[res_len]= '[';
          memcpy(&result[res_len + 1], path[i]->key, path[i]->key_len);
          result[res_len + 1 + path[i]->key_len]= ']';
          res_len+= (path[i]->key_len + 2);
        }
      }
      
      result[res_len]= '\0';
      
      return result;
    }
    else
    {
      return NULL;
    }
  }
  else
  {
    if (this->exists && 0 < this->path_len)
    {
      char* result= new char[strlen(doc)]; //change to actual path len
      int res_len= 0;
      for (int i= 0; i < path_len; i++)
      {
        char* temp_res= escape(path[i]->key, path[i]->key_len, (const char*) ":");
        memcpy(&result[res_len], temp_res, strlen(temp_res));
        result[res_len + strlen(temp_res)]= ':';
        res_len+= (strlen(temp_res) + 1);
        if (temp_res)
          delete[] temp_res;
      }
      
      result[res_len++]= ':';
      result[res_len]= '\0';
      
      return result;
    }
    else
    {
      return NULL;
    }
  }
};


JCallback_search::~JCallback_search()
{
   if (this->path)
    {
      for (int i= 0; i < path_len; i++)
        delete path[i];
      delete[] this->path;
    }
};


JCallback_depth::JCallback_depth()
{
  this->depth= 0;
};


bool JCallback_depth::callback_func(int key_start, int key_len, int val_start, int val_len, int pos, bool is_array_elem, int type_flag)
{
  if (pos > this->depth)
    this->depth= pos;
  
  return true;
};


int JCallback_depth::get_depth()
{
  return this->depth + 1; /* adding 1 here, because parser does not count root element */
}


JCallback_count::JCallback_count()
{
  this->count= 0;
};


bool JCallback_count::callback_func(int key_start, int key_len, int val_start, int val_len, int pos, bool is_array_elem, int type_flag)
{
  if (1 == pos)
    this->count++;
  
  return true;
};


int JCallback_count::get_count()
{
  return this->count; /* adding 1 here, because parser does not count root element */
}


JCallback_merge::JCallback_merge(char* doc)
{
  this->hier= -1;
  this->save_pos= -1;
  this->pos_in_cur= 0;
  this->root= NULL;
  this->doc= doc;
  this->garbage= new Jtree_elem*[MAX_JSON_DEPTH];
  this->garbage_pos= 0;
};


JCallback_merge::~JCallback_merge()
{
  clean_root(this->root);
  for (int i= 0; i < this->garbage_pos; i++)
    delete this->garbage[i];
  delete[] this->garbage;
  delete[] this->cur->cur;
  delete this->cur->parent;
  delete this->cur;
};


void JCallback_merge::clean_root(Jtree_elem* elem)
{
  if (NULL != elem)
  {
    if (0 < elem->num_childs)
    {
      for (int i= 0; i < elem->num_childs; i++)
      {
        clean_root(elem->childs[i]);
      }
      elem->num_childs= 0;
      if (elem->childs)
      {
        delete[] elem->childs;
      }
    }
  }
}


bool JCallback_merge::callback_func(int key_start, int key_len, int val_start, int val_len, int pos, bool is_array_elem, int type_flag)
{
  Jtree_elem* je= new Jtree_elem(this->jstack, this->doc);
  this->garbage[this->garbage_pos++]= je;
  je->pos_in_stack= num_elem;
  je->num_childs= 0;
  je->childs= NULL;
  if (0 > this->save_pos)
  {
    this->save_pos= pos;
    this->hier= 1;
    this->cur= new Jtree_elem_childs();
    this->cur->cur= new Jtree_elem*[MAX_JSON_DEPTH];
  }
  
  if (pos < save_pos)
  {
    hier++;
    je->childs= this->cur->cur;
    je->num_childs= this->cur->cur_pos;
    if (NULL != this->cur->parent)
    {
      Jtree_elem_childs* temp= this->cur->parent;
      delete this->cur;
      this->cur= temp;
    }
    else
    {
      delete this->cur;
      Jtree_elem_childs* temp= new Jtree_elem_childs();
      temp->cur= new Jtree_elem*[MAX_JSON_DEPTH];
      temp->parent= NULL;
      temp->cur_pos= 0;
      this->cur= temp;
    }
    this->save_pos= pos;
  }
  else if (pos > save_pos)
  {
    hier--;
    Jtree_elem_childs* temp= new Jtree_elem_childs();
    temp->cur= new Jtree_elem*[MAX_JSON_DEPTH];
    temp->parent= this->cur;
    temp->cur_pos= 0;
    this->cur= temp;
    this->save_pos= pos;
  }
  
  if (0 == pos)
  {
    this->root= je;
  }
  else
  {
    je->num_links++;
    this->cur->cur[this->cur->cur_pos]= je;
    this->cur->cur_pos++;
  }
  
  put_elem(key_start, key_len, val_start, val_len, type_flag);
  return true;
};


Jtree_elem* JCallback_merge::get_merge_tree()
{
  return this->root;
};


stack_jelem* JCallback_merge::get_stack()
{
  return this->jstack;
};

/* :tabSize=2:indentSize=2:noTabs=true:folding=custom:collapseFolds=1: */
