/* 
  Copyright (c) 2013 - 2014, Oracle and/or its affiliates. All rights reserved.
  
  This program is free software; you can redistribute it and/or modify it under 
  the terms of the GNU General Public License as published by the Free Software 
  Foundation; version 2 of the License.
  
  This program is distributed in the hope that it will be useful, but WITHOUT ANY 
  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS 
  FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
  
  You should have received a copy of the GNU General Public License along with this 
  program; if not, write to the Free Software Foundation, Inc., 51 Franklin St, Fifth 
  Floor, Boston, MA 02110-1301  USA
*/
#pragma once

#ifndef JSTACK
#define JSTACK

#include <stddef.h>
#include "jutil.h"

#define MAX_JSON_DEPTH 1024                   /* maximum depth of JSON document */

/* 
  Type of JSON element.
*/

enum enum_jtype
{
	jdefault= 0,
	jtrue=    1,
	jfalse=   2,
	jnull=    3,
	jnum=     4,
	jstring=  5,
	jarray=   6,
	jobject=  7,
	jpair=    8,
	jutf=     9
};

/* 
  Stack which holds information about found keys
*/

typedef struct sj {
  int key_start;                             /* Pointer to the key name in the doc */
  unsigned int key_len;                       /* Length of the key name */
  int val_start;                                /* Pointer to the value */
  unsigned int val_len;                       /* Length of the value */
  int parent;                                 /* Index of parent element in the stack */
  int type_flag;                       /* Type of the element */
} stack_jelem;


/* 
  Tree element, needed for JSON_MERGE
*/

class Jtree_elem {
public:
  int pos_in_stack;             /* Position of the element in jstack */
  int num_childs;               /* Number of childs */
  Jtree_elem** childs;          /* Childs array */
  
  stack_jelem* st;            /* Pointer to stack_jelem */
  char* doc;                  /* Pointer to doc */
  int val_offset;             /* Offset for length of value if changed */
  int val_start;              /* Initial value start */
  bool changed;               /* Was value changed or not? */
  bool is_new;                /* Is this new element? */
  int offset;                 /* Offset from key_start and val_start positions */
  int num_links;              /* Number of links */
  
  Jtree_elem(stack_jelem* st, char* doc)
  {
    this->st= st;
    this->doc= doc;
    val_offset= 0;
    changed= false;
    is_new= false;
    offset= 0;
    this->val_start= 0;
    num_childs= 0;
    num_links= 0;
  }
};


/* 
  Childs tree
*/
class Jtree_elem_childs
{
public:
  Jtree_elem_childs* parent;    /* Parent array */
  Jtree_elem** cur;             /* Current element in the tree */
  int cur_pos;                  /* Current position in the array */
  
  Jtree_elem_childs() { parent= NULL; cur_pos= 0; }
};


#endif /* JSTACK */

/*
* :tabSize=2:indentSize=2:noTabs=true:
* :folding=custom:collapseFolds=1:
*/
