/* 
  Copyright (c) 2013 - 2014, Oracle and/or its affiliates. All rights reserved.
  
  This program is free software; you can redistribute it and/or modify it under 
  the terms of the GNU General Public License as published by the Free Software 
  Foundation; version 2 of the License.
  
  This program is distributed in the hope that it will be useful, but WITHOUT ANY 
  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS 
  FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
  
  You should have received a copy of the GNU General Public License along with this 
  program; if not, write to the Free Software Foundation, Inc., 51 Franklin St, Fifth 
  Floor, Boston, MA 02110-1301  USA
*/

#include "jutil.h"

#include <stdarg.h>
#include <string.h>


void jdump(int level, int number, const char* fmt, ...)
{
  if (JSON_VERBOSE_LEVEL > level)
    return;
  char* message= new char[1024];
  va_list args;
  va_start(args,fmt);
  vsprintf(message, fmt, args);
  va_end(args);

  switch (number)
  {
  case JSON_PARSE_ERROR:
    fprintf(stderr, "Parse error: %s\n", message);
    break;
  case JSON_CALLBACK_ERROR:
    fprintf(stderr, "Processing error: %s\n", message);
    break;
  case JSON_FUNCTION_ERROR:
    fprintf(stderr, "Post-processing error: %s\n", message);
    break;
  case JSON_UDF_ERROR:
    fprintf(stderr, "UDF-level error: %s\n", message);
    break;
  default:
    fprintf(stderr, "Unknown error: %s\n", message);
    break;
  }
  
  delete[] message;
};


char* escape(const char* string, int slen, const char* what_to_escape)
{
  char* result= new char[slen * 2];
  int wtel= strlen(what_to_escape);
  int offset= 0;
  for (int i= 0; i < slen; i++)
  {
    for (int j= 0; j < wtel; j++)
    {
      if (string[i] == what_to_escape[j])
      {
        result[i + offset]= '\\';
        offset++;
      }
    }
    result[i + offset]= string[i];
  }
  
  result[slen + offset]= '\0';
  
  return result;
};
/*
* :tabSize=2:indentSize=2:noTabs=true:
* :folding=custom:collapseFolds=1:
*/

