/* 
  Copyright (c) 2014, Oracle and/or its affiliates. All rights reserved.
  
  This program is free software; you can redistribute it and/or modify it under 
  the terms of the GNU General Public License as published by the Free Software 
  Foundation; version 2 of the License.
  
  This program is distributed in the hope that it will be useful, but WITHOUT ANY 
  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS 
  FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
  
  You should have received a copy of the GNU General Public License along with this 
  program; if not, write to the Free Software Foundation, Inc., 51 Franklin St, Fifth 
  Floor, Boston, MA 02110-1301  USA
*/


/* 
  This file contains helper functions, used by whole JSON UDF package.
*/
#pragma once

#ifndef JUTIL
#define JUTIL


/* Error numbers */
#define JSON_SUCCESS        0
#define JSON_PARSE_ERROR    1
#define JSON_CALLBACK_ERROR 2
#define JSON_FUNCTION_ERROR 3
#define JSON_UDF_ERROR      4

/* Error levels */
#define JSON_NOTE     1
#define JSON_WARNING  2
#define JSON_ERROR    3

/* Current print level should be defined at compile time */
//#define JSON_VERBOSE_LEVEL  1

#include <stdio.h>
#include <stdlib.h>

#include "jkey.h"


/* Structure which holds JKey** array and size of it */
struct JKeyHolder
{
  JKey** array;     //array itself
  size_t size;      //size of the array
  
  JKeyHolder(JKey** a, size_t s): array(a), size(s) {};
  
  ~JKeyHolder()
  {
    for (unsigned int i= 0; i < size; i++)
    {
      delete array[i];
    }
    
    delete[] array;
  };
};


/* 
  Dumps a message into error log file
  
  @param[in]  level   error level
  @param[in]  number  error number
  @param[in]  fmt     format
  @param[in]  ...     optional arguments to pass to format string
*/
void jdump(int level, int number, const char* fmt, ...);


/* 
  Escapes special characters in the string
  
  @param[in]  string          source string
  @param[in]  slen            source string length
  @param[in]  what_to_escape  character array of characters to escape
  
  @return escaped string
*/
char* escape(const char* string, int slen, const char* what_to_escape);


#endif /* JUTIL */

/*
* :tabSize=2:indentSize=2:noTabs=true:
* :folding=custom:collapseFolds=1:
*/
