/* 
  Copyright (c) 2013 - 2014, Oracle and/or its affiliates. All rights reserved.
  
  This program is free software; you can redistribute it and/or modify it under 
  the terms of the GNU General Public License as published by the Free Software 
  Foundation; version 2 of the License.
  
  This program is distributed in the hope that it will be useful, but WITHOUT ANY 
  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS 
  FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
  
  You should have received a copy of the GNU General Public License along with this 
  program; if not, write to the Free Software Foundation, Inc., 51 Franklin St, Fifth 
  Floor, Boston, MA 02110-1301  USA
*/
#include "my_json_udf.h"

long long json_valid(UDF_INIT *initid, UDF_ARGS *args, char *is_null, char *error)
{
	//if args->args[0] is NULL, no need to parse
	//we can not return this at init state, because this is not an error
	if (!args->args[0])
	{
    jdump(JSON_NOTE, JSON_UDF_ERROR, "First argument is NULL");
		return false;
	}
	
	char *doc= new char[args->lengths[0] + 1];
  memcpy(doc, args->args[0], args->lengths[0]);
  doc[args->lengths[0]]= '\0';
  
	bool result= json_valid(doc);
	delete[] doc;
	if (result)
		return 1;
	else
		return 0;
}


my_bool json_valid_init(UDF_INIT *initid, UDF_ARGS *args, char *message)
{
	if (args->arg_count != 1 || args->arg_type[0] != STRING_RESULT) {
	  strcpy(message,"JSON_VALID only takes 1 string argument");
	  return 1;
	}

	initid->maybe_null = 1;
	initid->max_length = 1;
	initid->const_item = 0;
	
	return 0;
}


void json_valid_deinit(UDF_INIT *initid)
{
	return json_deinit(initid);
}


char* json_search(UDF_INIT *initid, UDF_ARGS *args, char *result, unsigned long *length, char *is_null, char *error)
{
	if (!args->args[0])
	{
    jdump(JSON_NOTE, JSON_UDF_ERROR, "First argument is NULL");
    return NULL;
  }
  
  if (!args->args[1] && 0 < args->lengths[1])
  {
    jdump(JSON_NOTE, JSON_UDF_ERROR, "Second argument is NULL, but args->lengths[1] is greater than 0, resetting");
    args->lengths[1]= 0;
  }
  
  int search_pos= 1;
  
	char *doc= new char[args->lengths[0] + 1];
  memcpy(doc, args->args[0], args->lengths[0]);
  doc[args->lengths[0]]= '\0';
  
	char* value= new char[args->lengths[1] + 1];
  memcpy(value, args->args[1], args->lengths[1]);
  value[args->lengths[1]]= '\0';
  
  if (3 == args->arg_count)
    search_pos= *((long long*) args->args[2]);
  
	char* result_string= json_search(doc, value, search_pos);

	delete[] doc;
	delete[] value;
	
  if (result_string)
  {
    if (MAXRES < strlen(result_string))
    {
      free(initid->ptr);
      if (!(initid->ptr= (char*)malloc(strlen(result_string))))
      {
        /* malloc error, no way to return proper error */
        delete[] result_string;
        jdump(JSON_ERROR, JSON_UDF_ERROR, "Cannot allocate memory for result string");
        return NULL;
      }
    }
    memcpy(initid->ptr, result_string, strlen(result_string));
    *length= (unsigned long)strlen(result_string);
    delete[] result_string;
    
    return initid->ptr;
  }
  else
    return NULL;
}


my_bool json_search_init(UDF_INIT *initid, UDF_ARGS *args, char *message)
{
	if (args->arg_count < 2 || args->arg_count > 3)
  {
    strcpy(message,"You must supply 2 required string arguments and 1 optional integer argument");
    return 1;
  }

  for (unsigned int i= 0; i < 2; i++)
  {
    if (args->arg_type[i] != STRING_RESULT)
    {   
      strcpy(message,"First two arguments for JSON_SEARCH must be strings");
      return 1;
    }   
  }
  
  if (3 == args->arg_count && args->arg_type[2] != INT_RESULT)
  {
    strcpy(message,"Optional argument to JSON_SEARCH must be integer");
    return 1;
  }

  if (!(initid->ptr= (char*)malloc(MAXRES)))
  {
    strcpy(message,"Couldn't allocate memory for result buffer");
    return 1;
  }
  memset(initid->ptr, '\0', MAXRES);
  
  initid->maybe_null= 1;
  initid->const_item= 0;
	
	return 0;
}


void json_search_deinit(UDF_INIT *initid)
{
	return json_deinit(initid);
}


char* json_extract(UDF_INIT *initid, UDF_ARGS *args, char *result, unsigned long *length, char *is_null, char *error)
{
	if (!args->args[0])
	{
    jdump(JSON_NOTE, JSON_UDF_ERROR, "First argument is NULL");
    return NULL;
  }
  
	char *doc= new char[args->lengths[0] + 1];
  memcpy(doc, args->args[0], args->lengths[0]);
  doc[args->lengths[0]]= '\0';
  
  JKey** jeargs= make_jeargs(args, 1, args->arg_count - 1);
  
	char* result_string= json_extract(doc, jeargs, args->arg_count - 1);
	
	for (unsigned int i= 0; i < args->arg_count - 1; i++)
  {
    delete jeargs[i];
  }

  delete[] jeargs;
  delete[] doc;

  if (result_string)
  {
    if (MAXRES < strlen(result_string))
    {
      free(initid->ptr);
      if (!(initid->ptr= (char*)malloc(strlen(result_string))))
      {
        /* malloc error, no way to return proper error */
        delete[] result_string;
        jdump(JSON_ERROR, JSON_UDF_ERROR, "Cannot allocate memory for result string");
        return NULL;
      }
    }
    memcpy(initid->ptr, result_string, strlen(result_string));
    *length= (unsigned long)strlen(result_string);
    delete[] result_string;
    
    return initid->ptr;
  }
  else
    return NULL;
}


my_bool json_extract_init(UDF_INIT *initid, UDF_ARGS *args, char *message)
{
	if (args->arg_count < 2)
  {
    strcpy(message,"You must supply at least 2 arguments");
    return 1;
  }

  for (unsigned int i= 0; i < args->arg_count; i++)
  {
    if (args->arg_type[i] != STRING_RESULT 
      && args->arg_type[i] != INT_RESULT
      )
    {
      strcpy(message,"JSON_EXTRACT accepts only string and integer arguments");
      return 1;
    }   
  }

  if (!(initid->ptr= (char*)malloc(MAXRES)))
  {
    strcpy(message,"Couldn't allocate memory for result buffer");
    return 1;
  }
  memset(initid->ptr, '\0', MAXRES);
  
  initid->maybe_null= 1;
  initid->const_item= 0;
	
	return 0;
}


void json_extract_deinit(UDF_INIT *initid)
{
	return json_deinit(initid);
}


char* json_replace(UDF_INIT *initid, UDF_ARGS *args, char *result, unsigned long *length, char *is_null, char *error)
{
	if (!args->args[0])
	{
    jdump(JSON_NOTE, JSON_UDF_ERROR, "First argument is NULL");
    return NULL;
  }
  
	char* doc= new char[args->lengths[0] + 1];
  memcpy(doc, args->args[0], args->lengths[0]);
  doc[args->lengths[0]]= '\0';
  
  JKey** jeargs= make_jeargs(args, 1, args->arg_count - 2);
  
  char* new_value= create_new_value(args, args->arg_count - 1);
  
  if (!new_value)
  {
    jdump(JSON_ERROR, JSON_UDF_ERROR, "Cannot create new value for JSON_REPLACE");
    delete[] jeargs;
    delete[] doc;
    return NULL;
  }

	char* result_string= json_replace(doc, jeargs, args->arg_count - 2, new_value);
	
	for (unsigned int i= 0; i < args->arg_count - 2; i++)
  {
    delete jeargs[i];
  }

  delete[] jeargs;
  delete[] doc;
  delete[] new_value;

  if (result_string)
  {
    if (MAXRES < strlen(result_string))
    {
      free(initid->ptr);
      if (!(initid->ptr= (char*)malloc(strlen(result_string))))
      {
        /* malloc error, no way to return proper error */
        delete[] result_string;
        jdump(JSON_ERROR, JSON_UDF_ERROR, "Cannot allocate memory for result string");
        return NULL;
      }
    }
    memcpy(initid->ptr, result_string, strlen(result_string));
    *length= (unsigned long)strlen(result_string);
    delete[] result_string;
    
    return initid->ptr;
  }
  else
    return NULL;
}


my_bool json_replace_init(UDF_INIT *initid, UDF_ARGS *args, char *message)
{
	if (args->arg_count < 3)
  {
    strcpy(message,"You must supply at least 3 arguments");
    return 1;
  }

  for (unsigned int i= 0; i < args->arg_count - 1; i++)
  {
    if (args->arg_type[i] != STRING_RESULT 
      && args->arg_type[i] != INT_RESULT
      )
    {   
      strcpy(message,"JSON_REPLACE accepts only string and integer arguments");
      return 1;
    }   
  }

  if (!(initid->ptr= (char*)malloc(MAXRES)))
  {
    strcpy(message,"Couldn't allocate memory for result buffer");
    return 1;
  }
  memset(initid->ptr, '\0', MAXRES);
  
  initid->maybe_null= 1;
  initid->const_item= 0;
	
	return 0;
}


void json_replace_deinit(UDF_INIT *initid)
{
	return json_deinit(initid);
}


char* json_append(UDF_INIT *initid, UDF_ARGS *args, char *result, unsigned long *length, char *is_null, char *error)
{
	if (!args->args[0])
	{
    jdump(JSON_NOTE, JSON_UDF_ERROR, "First argument is NULL");
    return NULL;
  }
  
	char* doc= new char[args->lengths[0] + 1];
  memcpy(doc, args->args[0], args->lengths[0]);
  doc[args->lengths[0]]= '\0';
  
  JKey** jeargs= make_jeargs(args, 1, args->arg_count - 2);
  
  char* new_value= create_new_value(args, args->arg_count - 1);
  
  if (!new_value)
  {
    jdump(JSON_ERROR, JSON_UDF_ERROR, "Cannot create new value for JSON_APPEND");
    delete[] jeargs;
    delete[] doc;
    return NULL;
  }

	char* result_string= json_append(doc, jeargs, args->arg_count - 2, new_value);
	
	for (unsigned int i= 0; i < args->arg_count - 2; i++)
  {
    delete jeargs[i];
  }

  delete[] jeargs;
  delete[] doc;
  delete[] new_value;

  if (result_string)
  {
    if (MAXRES < strlen(result_string))
    {
      free(initid->ptr);
      if (!(initid->ptr= (char*)malloc(strlen(result_string))))
      {
        /* malloc error, no way to return proper error */
        delete[] result_string;
        jdump(JSON_ERROR, JSON_UDF_ERROR, "Cannot allocate memory for result string");
        return NULL;
      }
    }
    memcpy(initid->ptr, result_string, strlen(result_string));
    *length= (unsigned long)strlen(result_string);
    delete[] result_string;
    
    return initid->ptr;
  }
  else
    return NULL;
}


my_bool json_append_init(UDF_INIT *initid, UDF_ARGS *args, char *message)
{
	if (args->arg_count < 3)
  {
    strcpy(message,"You must supply at least 3 arguments");
    return 1;
  }

  for (unsigned int i= 0; i < args->arg_count - 1; i++)
  {
    if (args->arg_type[i] != STRING_RESULT 
      && args->arg_type[i] != INT_RESULT
      )
    {   
      strcpy(message,"JSON_APPEND accepts only string and integer arguments");
      return 1;
    }   
  }
  
  if (!(initid->ptr= (char*)malloc(MAXRES)))
  {
    strcpy(message,"Couldn't allocate memory for result buffer");
    return 1;
  }
  memset(initid->ptr, '\0', MAXRES);
  
  initid->maybe_null= 1;
  initid->const_item= 0;
	
	return 0;
}


void json_append_deinit(UDF_INIT *initid)
{
	return json_deinit(initid);
	
}


char* json_remove(UDF_INIT *initid, UDF_ARGS *args, char *result, unsigned long *length, char *is_null, char *error)
{
	if (!args->args[0])
	{
    jdump(JSON_NOTE, JSON_UDF_ERROR, "First argument is NULL");
    return NULL;
  }
  
	char* doc= new char[args->lengths[0] + 1];
  memcpy(doc, args->args[0], args->lengths[0]);
  doc[args->lengths[0]]= '\0';
  
  JKey** jeargs= make_jeargs(args, 1, args->arg_count - 1);
  
	char* result_string= json_remove(doc, jeargs, args->arg_count - 1);
	
	for (unsigned int i= 0; i < args->arg_count - 1; i++)
  {
    delete jeargs[i];
  }

  delete[] jeargs;
  delete[] doc;

  if (result_string)
  {
    if (MAXRES < strlen(result_string))
    {
      free(initid->ptr);
      if (!(initid->ptr= (char*)malloc(strlen(result_string))))
      {
        /* malloc error, no way to return proper error */
        delete[] result_string;
        jdump(JSON_ERROR, JSON_UDF_ERROR, "Cannot allocate memory for result string");
        return NULL;
      }
    }
    memcpy(initid->ptr, result_string, strlen(result_string));
    *length= (unsigned long)strlen(result_string);
    delete[] result_string;
    
    return initid->ptr;
  }
  else
    return NULL;
}


my_bool json_remove_init(UDF_INIT *initid, UDF_ARGS *args, char *message)
{
	if (args->arg_count < 2)
  {
    strcpy(message,"You must supply at least 2 arguments");
    return 1;
  }

  for (unsigned int i= 0; i < args->arg_count; i++)
  {
    if (args->arg_type[i] != STRING_RESULT 
      && args->arg_type[i] != INT_RESULT
      )
    {   
      strcpy(message,"JSON_REMOVE accepts only string and integer arguments");
      return 1;
    }   
  }

  if (!(initid->ptr= (char*)malloc(MAXRES)))
  {
    strcpy(message,"Couldn't allocate memory for result buffer");
    return 1;
  }
  memset(initid->ptr, '\0', MAXRES);
  
  initid->maybe_null= 1;
  initid->const_item= 0;
	
	return 0;
}


void json_remove_deinit(UDF_INIT *initid)
{
	return json_deinit(initid);
}


char* json_set(UDF_INIT *initid, UDF_ARGS *args, char *result, unsigned long *length, char *is_null, char *error)
{
	if (!args->args[0])
	{
    jdump(JSON_NOTE, JSON_UDF_ERROR, "First argument is NULL");
    return NULL;
  }
  
	char* doc= new char[args->lengths[0] + 1];
  memcpy(doc, args->args[0], args->lengths[0]);
  doc[args->lengths[0]]= '\0';
  
  JKey** jeargs= make_jeargs(args, 1, args->arg_count - 2);
  
  char* new_value= create_new_value(args, args->arg_count - 1);
  
  if (!new_value)
  {
    jdump(JSON_ERROR, JSON_UDF_ERROR, "Cannot create new value for JSON_SET");
    delete[] jeargs;
    delete[] doc;
    return NULL;
  }

	char* result_string= json_set(doc, jeargs, args->arg_count - 2, new_value);
	
	for (unsigned int i= 0; i < args->arg_count - 2; i++)
  {
    delete jeargs[i];
  }

  delete[] jeargs;
  delete[] doc;
  delete[] new_value;

  if (result_string)
  {
    if (MAXRES < strlen(result_string))
    {
      free(initid->ptr);
      if (!(initid->ptr= (char*)malloc(strlen(result_string))))
      {
        /* malloc error, no way to return proper error */
        delete[] result_string;
        jdump(JSON_ERROR, JSON_UDF_ERROR, "Cannot allocate memory for result string");
        return NULL;
      }
    }
    memcpy(initid->ptr, result_string, strlen(result_string));
    *length= (unsigned long)strlen(result_string);
    delete[] result_string;
    
    return initid->ptr;
  }
  else
    return NULL;
}


my_bool json_set_init(UDF_INIT *initid, UDF_ARGS *args, char *message)
{
	if (args->arg_count < 3)
  {
    strcpy(message,"You must supply at least 3 arguments");
    return 1;
  }

  for (unsigned int i= 0; i < args->arg_count - 1; i++)
  {
    if (args->arg_type[i] != STRING_RESULT 
      && args->arg_type[i] != INT_RESULT
      )
    {   
      strcpy(message,"JSON_SET accepts only string and integer arguments");
      return 1;
    }   
  }

  if (!(initid->ptr= (char*)malloc(MAXRES)))
  {
    strcpy(message,"Couldn't allocate memory for result buffer");
    return 1;
  }
  memset(initid->ptr, '\0', MAXRES);
  
  initid->maybe_null= 1;
  initid->const_item= 0;
	
	return 0;
}


void json_set_deinit(UDF_INIT *initid)
{
	return json_deinit(initid);
}


char* json_merge(UDF_INIT *initid, UDF_ARGS *args, char *result, unsigned long *length, char *is_null, char *error)
{
	return json_merge_do(initid, args, result, length, is_null, error);
}


my_bool json_merge_init(UDF_INIT *initid, UDF_ARGS *args, char *message)
{
	return json_merge_do_init(initid, args, message);
}


void json_merge_deinit(UDF_INIT *initid)
{
	return json_deinit(initid);
}


char* json_safe_merge(UDF_INIT *initid, UDF_ARGS *args, char *result, unsigned long *length, char *is_null, char *error)
{
	return json_merge_do(initid, args, result, length, is_null, error, true);
}


my_bool json_safe_merge_init(UDF_INIT *initid, UDF_ARGS *args, char *message)
{
	return json_merge_do_init(initid, args, message, true);
}


void json_safe_merge_deinit(UDF_INIT *initid)
{
	return json_deinit(initid);
}


char* json_deep_merge(UDF_INIT *initid, UDF_ARGS *args, char *result, unsigned long *length, char *is_null, char *error)
{
	return json_merge_do(initid, args, result, length, is_null, error, false, true);
}


my_bool json_deep_merge_init(UDF_INIT *initid, UDF_ARGS *args, char *message)
{
	return json_merge_do_init(initid, args, message, true, true);
}


void json_deep_merge_deinit(UDF_INIT *initid)
{
	return json_deinit(initid);
}


long long json_contains_key(UDF_INIT *initid, UDF_ARGS *args, char *is_null, char *error)
{
  if (!args->args[0])
  {
    jdump(JSON_NOTE, JSON_UDF_ERROR, "First argument is NULL");
    return false;
  }
  
	char *doc= new char[args->lengths[0] + 1];
  memcpy(doc, args->args[0], args->lengths[0]);
  doc[args->lengths[0]]= '\0';
  
  JKey** jeargs= make_jeargs(args, 1, args->arg_count - 1);
  
	bool result= json_contains_key(doc, jeargs, args->arg_count - 1);
	
	for (unsigned int i= 0; i < args->arg_count - 1; i++)
  {
    delete jeargs[i];
  }

  delete[] jeargs;
  delete[] doc;

  if (result)
    return 1;
  else
    return 0;
}


my_bool json_contains_key_init(UDF_INIT *initid, UDF_ARGS *args, char *message)
{
  if (args->arg_count < 2)
  {
    strcpy(message,"You must supply at least 2 arguments");
    return 1;
  }

  for (unsigned int i= 0; i < args->arg_count; i++)
  {
    if (args->arg_type[i] != STRING_RESULT 
      && args->arg_type[i] != INT_RESULT
      )
    {   
      strcpy(message,"JSON_CONTAINS_KEY accepts only string and integer arguments");
      return 1;
    }   
  }

  initid->maybe_null= 1;
  initid->const_item= 0;
	
	return 0;
}


void json_contains_key_deinit(UDF_INIT *initid)
{
	return json_deinit(initid);	
}

char* json_test_parser(UDF_INIT *initid, UDF_ARGS *args, char *result, unsigned long *length, char *is_null, char *error)
{
  if (NULL == args->args[0] && 0 != args->lengths[0])
  {
    jdump(JSON_ERROR, JSON_UDF_ERROR, "Broken argument received");
    return NULL;
  }
  
  char *doc= new char[args->lengths[0] + 1];
  memcpy(doc, args->args[0], args->lengths[0]);
  doc[args->lengths[0]]= '\0';
  
  char* result_string= json_test_parser(doc);
  delete[] doc;
  
  if (result_string)
  {
    if (MAXRES < strlen(result_string))
    {
      free(initid->ptr);
      if (!(initid->ptr= (char*)malloc(strlen(result_string))))
      {
        /* malloc error, no way to return proper error */
        delete[] result_string;
        jdump(JSON_ERROR, JSON_UDF_ERROR, "Cannot allocate memory for result string");
        return NULL;
      }
    }
    memcpy(initid->ptr, result_string, strlen(result_string));
    *length= (unsigned long)strlen(result_string);
    delete[] result_string;
    
    return initid->ptr;
  }
  else
  {
    delete[] result_string;
    return NULL;
  }
}


my_bool json_test_parser_init(UDF_INIT *initid, UDF_ARGS *args, char *message)
{
	if (args->arg_count != 1 || args->arg_type[0] != STRING_RESULT) {
	  strcpy(message,"JSON_TEST_PARSER only takes 1 string argument\n");
	  return 1;
	}

  if (!(initid->ptr= (char*)malloc(MAXRES)))
  {
    strcpy(message,"Couldn't allocate memory for result buffer for JSON_TEST_PARSER\n");
    return 1;
  }
  memset(initid->ptr, '\0', MAXRES);

	initid->maybe_null = 1;
	initid->max_length = 1;
	initid->const_item = 0;
	
	return 0;
}


void json_test_parser_deinit(UDF_INIT *initid)
{
	return json_deinit(initid);
}


long long json_depth(UDF_INIT *initid, UDF_ARGS *args, char *is_null, char *error)
{
	//if args->args[0] is NULL, no need to parse
	//we can not return this at init state, because this is not an error
	if (!args->args[0])
	{
    jdump(JSON_NOTE, JSON_UDF_ERROR, "First argument is NULL");
		return false;
	}
	
	char *doc= new char[args->lengths[0] + 1];
  memcpy(doc, args->args[0], args->lengths[0]);
  doc[args->lengths[0]]= '\0';
  
	int result= json_depth(doc);
	delete[] doc;
	
	return result;
}


my_bool json_depth_init(UDF_INIT *initid, UDF_ARGS *args, char *message)
{
	if (args->arg_count != 1 || args->arg_type[0] != STRING_RESULT) {
	  strcpy(message,"JSON_DEPTH only takes 1 string argument");
	  return 1;
	}

	initid->maybe_null = 1;
	initid->max_length = 1;
	initid->const_item = 0;
	
	return 0;
}


void json_depth_deinit(UDF_INIT *initid)
{
	return json_deinit(initid);
}

long long json_count(UDF_INIT *initid, UDF_ARGS *args, char *is_null, char *error)
{
	//if args->args[0] is NULL, no need to parse
	//we can not return this at init state, because this is not an error
	if (!args->args[0])
	{
    jdump(JSON_NOTE, JSON_UDF_ERROR, "First argument is NULL");
		return false;
	}
	
	char *doc= new char[args->lengths[0] + 1];
  memcpy(doc, args->args[0], args->lengths[0]);
  doc[args->lengths[0]]= '\0';
  
  JKey** jeargs= make_jeargs(args, 1, args->arg_count - 1);
  
	int result= json_count(doc, jeargs, args->arg_count - 1);
	
	for (unsigned int i= 0; i < args->arg_count - 1; i++)
  {
    delete jeargs[i];
  }

	delete[] jeargs;
	delete[] doc;
	
	return result;
}


my_bool json_count_init(UDF_INIT *initid, UDF_ARGS *args, char *message)
{
	if (args->arg_count < 1) {
	  strcpy(message,"You must supply at least 1 argument");
	  return 1;
	}
	
  for (unsigned int i= 0; i < args->arg_count; i++)
  {
    if (args->arg_type[i] != STRING_RESULT 
      && args->arg_type[i] != INT_RESULT
      )
    {   
      strcpy(message,"JSON_COUNT accepts only string and integer arguments");
      return 1;
    }   
  }

	initid->maybe_null = 1;
	initid->max_length = 1;
	initid->const_item = 0;
	
	return 0;
}


void json_count_deinit(UDF_INIT *initid)
{
	return json_deinit(initid);
}



char* json_version(UDF_INIT *initid, UDF_ARGS *args, char *result, unsigned long *length, char *is_null, char *error)
{
  char* result_string= json_version();
  
  if (result_string)
  {
    if (MAXRES < strlen(result_string))
    {
      free(initid->ptr);
      if (!(initid->ptr= (char*)malloc(strlen(result_string))))
      {
        /* malloc error, no way to return proper error */
        delete[] result_string;
        jdump(JSON_ERROR, JSON_UDF_ERROR, "Cannot allocate memory for result string");
        return NULL;
      }
    }
    memcpy(initid->ptr, result_string, strlen(result_string));
    *length= (unsigned long)strlen(result_string);
    delete[] result_string;
    
    return initid->ptr;
  }
  else
  {
    delete[] result_string;
    return NULL;
  }
}


my_bool json_version_init(UDF_INIT *initid, UDF_ARGS *args, char *message)
{
	if (args->arg_count != 0) {
	  strcpy(message,"JSON_VERSION takes no argument\n");
	  return 1;
	}

  if (!(initid->ptr= (char*)malloc(MAXRES)))
  {
    strcpy(message,"Couldn't allocate memory for result buffer for JSON_VERSION\n");
    return 1;
  }
  memset(initid->ptr, '\0', MAXRES);

	initid->maybe_null = 0;
	initid->const_item = 0;
	
	return 0;
}


void json_version_deinit(UDF_INIT *initid)
{
	return json_deinit(initid);
}


//helpers

void json_deinit(UDF_INIT *initid)
{
	if (initid->ptr)
		free(initid->ptr);
}


JKey** make_jeargs(UDF_ARGS *args, unsigned int start, unsigned int length)
{
  JKey** retval= new JKey*[length];
  for (unsigned int i= start; i < start + length; i++)
  {
    if (NULL == args->args[i])
    {
      jdump(JSON_ERROR, JSON_UDF_ERROR, "Argument %d is NULL", i);
      char* value= new char[1];
      value[0]= '\0';
      retval[i - 1]= new JKey_string(value);
    }
    else if (STRING_RESULT == args->arg_type[i])
    {
      char* value= new char[args->lengths[i] + 1];
      memcpy(value, args->args[i], args->lengths[i]);
      value[args->lengths[i]]= '\0';
      retval[i - 1]= new JKey_string(value);
    }
    else if (INT_RESULT == args->arg_type[i])
    {
      jdump(JSON_NOTE, JSON_UDF_ERROR, "int arg");
      retval[i - 1]= new JKey_longlong(*((long long*) args->args[i]));
    }
  }
  
  return retval;
}


char* json_merge_do(UDF_INIT *initid, UDF_ARGS *args, char *result, unsigned long *length, char *is_null, char *error, bool safe, bool deep)
{
  char* result_string;
  
	if (!args->args[0])
	{
    return NULL;
  }
  
  char** jeargs=new char*[args->arg_count];
  for (unsigned int i= 0; i < args->arg_count; i++)
  {
    if (!args->args[i] && 0 < args->lengths[i])
    {
      jdump(JSON_NOTE, JSON_UDF_ERROR, "Argument %d is NULL, but args->lengths[%d] is greater than 0, resetting", i, i);
      args->lengths[i]= 0;
    }
    jeargs[i]= new char[args->lengths[i] + 1];
    memcpy(jeargs[i], args->args[i], args->lengths[i]);
    jeargs[i][args->lengths[i]]= '\0';
  }

  result_string= json_merge(jeargs, args->arg_count, safe, deep);
	
	for (unsigned int i= 0; i < args->arg_count; i++)
  {
    delete[] jeargs[i];
  }

  delete[] jeargs;

  if (result_string)
  {
    if (MAXRES < strlen(result_string))
    {
      free(initid->ptr);
      if (!(initid->ptr= (char*)malloc(strlen(result_string))))
      {
        /* malloc error, no way to return proper error */
        delete[] result_string;
        jdump(JSON_ERROR, JSON_UDF_ERROR, "Cannot allocate memory for result string");
        return NULL;
      }
    }
    memcpy(initid->ptr, result_string, strlen(result_string));
    *length= (unsigned long)strlen(result_string);
    delete[] result_string;
    
    return initid->ptr;
  }
  else
  {
    return NULL;
  }
}


my_bool json_merge_do_init(UDF_INIT *initid, UDF_ARGS *args, char *message, bool safe, bool deep)
{
	if (args->arg_count < 2)
  {
    strcpy(message,"You must supply at least 2 string arguments");
    return 1;
  }

  for (unsigned int i= 0; i < args->arg_count; i++)
  {
    if (args->arg_type[i] != STRING_RESULT)
    {
      if (deep)
        strcpy(message,"JSON_DEEP_MERGE accepts only string arguments");
      else if (safe)
        strcpy(message,"JSON_SAFE_MERGE accepts only string arguments");
      else
        strcpy(message,"JSON_MERGE accepts only string arguments");
      return 1;
    }   
  }

  if (!(initid->ptr= (char*)malloc(MAXRES)))
  {
    free(initid->ptr);
    strcpy(message,"Couldn't allocate memory for result buffer");
    return 1;
  }
  memset(initid->ptr, '\0', MAXRES);
  
  initid->maybe_null= 1;
  initid->const_item= 0;
	
	return 0;
}

char* create_new_value(UDF_ARGS *args, int pos)
{
  char* new_value;
  if (!args->args[pos] && 0 < args->lengths[pos])
  {
    jdump(JSON_NOTE, JSON_UDF_ERROR, "In create_new_value: argument %d is NULL, but args->lengths[%d] is greater than 0, resetting", pos, pos);
    args->lengths[pos]= 0;
    args->arg_type[pos]= STRING_RESULT; /* Need to reset type as well to avoid rewriting handling of numbers */
  }
  
  switch(args->arg_type[pos])
  {
  case STRING_RESULT:
  case DECIMAL_RESULT:
    if (0 == args->lengths[pos])
    {
      if (args->args[pos])
      {
        new_value= new char[3];
        memcpy(new_value, "\"\"\0", 3);
      }
      else
      {
        new_value= new char[5];
        memcpy(new_value, "null\0", 5);
      }
    }
    else
    {
      new_value= new char[args->lengths[pos] + 1]; 
      memcpy(new_value, args->args[pos], args->lengths[pos]);
      new_value[args->lengths[pos]]= '\0';
    }
    break;
  case INT_RESULT:
    new_value= new char[MAX_BIGINT_WIDTH + 1];
    
    #ifdef __WIN__
      _set_output_format(_TWO_DIGIT_EXPONENT);
    #endif
    
    sprintf(new_value, "%lld",  *((long long*) args->args[pos]));
    break;
  case REAL_RESULT:
    new_value= new char[36 + 1];
    sprintf(new_value, "%e", *((double*) args->args[pos]));
    break;
  default:
    jdump(JSON_NOTE, JSON_UDF_ERROR, "Unknown type for new value");
    return NULL;
  }
  return new_value;
}
/*
* :tabSize=2:indentSize=2:noTabs=true:
* :folding=custom:collapseFolds=1:
*/
