/* 
  Copyright (c) 2013 - 2014, Oracle and/or its affiliates. All rights reserved.
  
  This program is free software; you can redistribute it and/or modify it under 
  the terms of the GNU General Public License as published by the Free Software 
  Foundation; version 2 of the License.
  
  This program is distributed in the hope that it will be useful, but WITHOUT ANY 
  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS 
  FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
  
  You should have received a copy of the GNU General Public License along with this 
  program; if not, write to the Free Software Foundation, Inc., 51 Franklin St, Fifth 
  Floor, Boston, MA 02110-1301  USA
*/
#pragma once

#ifndef MY_JSON_UDF
#define MY_JSON_UDF

#include <my_global.h>
#include <my_sys.h>
#include <mysql.h>

#include <string>
#include <vector>

//#include "JParser.h"
#include "my_json.h"
#include "jkey.h"


#ifdef __WIN__
#define WINEXPORT __declspec(dllexport)
//#define strcpy strcpy_s
#else
#define WINEXPORT
#endif

#define MAXRES 1024

extern "C" {

WINEXPORT long long json_valid(UDF_INIT *initid, UDF_ARGS *args, char *is_null, char *error);
WINEXPORT my_bool json_valid_init(UDF_INIT *initid, UDF_ARGS *args, char *message);
WINEXPORT void json_valid_deinit(UDF_INIT *initid);

WINEXPORT char* json_search(UDF_INIT *initid, UDF_ARGS *args, char *result, unsigned long *length, char *is_null, char *error);
WINEXPORT my_bool json_search_init(UDF_INIT *initid, UDF_ARGS *args, char *message);
WINEXPORT void json_search_deinit(UDF_INIT *initid);

WINEXPORT char* json_extract(UDF_INIT *initid, UDF_ARGS *args, char *result, unsigned long *length, char *is_null, char *error);
WINEXPORT my_bool json_extract_init(UDF_INIT *initid, UDF_ARGS *args, char *message);
WINEXPORT void json_extract_deinit(UDF_INIT *initid);

WINEXPORT char* json_replace(UDF_INIT *initid, UDF_ARGS *args, char *result, unsigned long *length, char *is_null, char *error);
WINEXPORT my_bool json_replace_init(UDF_INIT *initid, UDF_ARGS *args, char *message);
WINEXPORT void json_replace_deinit(UDF_INIT *initid);

WINEXPORT char* json_append(UDF_INIT *initid, UDF_ARGS *args, char *result, unsigned long *length, char *is_null, char *error);
WINEXPORT my_bool json_append_init(UDF_INIT *initid, UDF_ARGS *args, char *message);
WINEXPORT void json_append_deinit(UDF_INIT *initid);

WINEXPORT char* json_remove(UDF_INIT *initid, UDF_ARGS *args, char *result, unsigned long *length, char *is_null, char *error);
WINEXPORT my_bool json_remove_init(UDF_INIT *initid, UDF_ARGS *args, char *message);
WINEXPORT void json_remove_deinit(UDF_INIT *initid);

WINEXPORT char* json_set(UDF_INIT *initid, UDF_ARGS *args, char *result, unsigned long *length, char *is_null, char *error);
WINEXPORT my_bool json_set_init(UDF_INIT *initid, UDF_ARGS *args, char *message);
WINEXPORT void json_set_deinit(UDF_INIT *initid);

WINEXPORT char* json_merge(UDF_INIT *initid, UDF_ARGS *args, char *result, unsigned long *length, char *is_null, char *error);
WINEXPORT my_bool json_merge_init(UDF_INIT *initid, UDF_ARGS *args, char *message);
WINEXPORT void json_merge_deinit(UDF_INIT *initid);

WINEXPORT char* json_safe_merge(UDF_INIT *initid, UDF_ARGS *args, char *result, unsigned long *length, char *is_null, char *error);
WINEXPORT my_bool json_safe_merge_init(UDF_INIT *initid, UDF_ARGS *args, char *message);
WINEXPORT void json_safe_merge_deinit(UDF_INIT *initid);

WINEXPORT char* json_deep_merge(UDF_INIT *initid, UDF_ARGS *args, char *result, unsigned long *length, char *is_null, char *error);
WINEXPORT my_bool json_deep_merge_init(UDF_INIT *initid, UDF_ARGS *args, char *message);
WINEXPORT void json_deep_merge_deinit(UDF_INIT *initid);

WINEXPORT long long json_contains_key(UDF_INIT *initid, UDF_ARGS *args, char *is_null, char *error);
WINEXPORT my_bool json_contains_key_init(UDF_INIT *initid, UDF_ARGS *args, char *message);
WINEXPORT void json_contains_key_deinit(UDF_INIT *initid);

WINEXPORT char* json_test_parser(UDF_INIT *initid, UDF_ARGS *args, char *result, unsigned long *length, char *is_null, char *error);
WINEXPORT my_bool json_test_parser_init(UDF_INIT *initid, UDF_ARGS *args, char *message);
WINEXPORT void json_test_parser_deinit(UDF_INIT *initid);

WINEXPORT long long json_depth(UDF_INIT *initid, UDF_ARGS *args, char *is_null, char *error);
WINEXPORT my_bool json_depth_init(UDF_INIT *initid, UDF_ARGS *args, char *message);
WINEXPORT void json_depth_deinit(UDF_INIT *initid);

WINEXPORT long long json_count(UDF_INIT *initid, UDF_ARGS *args, char *is_null, char *error);
WINEXPORT my_bool json_count_init(UDF_INIT *initid, UDF_ARGS *args, char *message);
WINEXPORT void json_count_deinit(UDF_INIT *initid);

WINEXPORT char* json_version(UDF_INIT *initid, UDF_ARGS *args, char *result, unsigned long *length, char *is_null, char *error);
WINEXPORT my_bool json_version_init(UDF_INIT *initid, UDF_ARGS *args, char *message);
WINEXPORT void json_version_deinit(UDF_INIT *initid);

}

//functions not available for MySQL user.

//general deinit
/* WINEXPORT  */void json_deinit(UDF_INIT *initid);


/*
  Creates jeargs array for functions which search by key
  
  @param[in]  args    UDF function arguments
  @param[in]  start   start element
  @param[in]  length  how many elements to copy
  
  @return             JKey** array of key elements
  @retval             JKey** array of key elements
*/
JKey** make_jeargs(UDF_ARGS *args, unsigned int start, unsigned int length);


/*
  Creates jeargs array from path string for functions which search by key
  
  @param[in]  args    path to the element
  @param[in]  length  length of the path
  
  @return             JKeyHolder array of key elements with its size
  @retval             JKeyHolder array of key elements with its size
*/
JKeyHolder* make_jeargs_from_path(char *path, unsigned int length);


/* 
  Helper to avoid copy-paste when calling two almost identical functions:
  json_merge and json_safe_merge
*/
char* json_merge_do(UDF_INIT *initid, UDF_ARGS *args, char *result, unsigned long *length, char *is_null, char *error, bool safe= false, bool deep= false);
my_bool json_merge_do_init(UDF_INIT *initid, UDF_ARGS *args, char *message, bool safe= false, bool deep= false);


/* 
  Prepare new value for inserting in json_replace, json_append and json_set
  
  @param[in]  args    UDF function arguments
  @param[in]  pos     Element position
  
  @return             Element, prepared for insertion
  @retval             char*
*/
char* create_new_value(UDF_ARGS *args, int pos);

#endif /* MY_JSON_UDF */

/* :tabSize=2:indentSize=2:noTabs=true:folding=custom:collapseFolds=1: */
