/* 
  Copyright (c) 2013 - 2014, Oracle and/or its affiliates. All rights reserved.
  
  This program is free software; you can redistribute it and/or modify it under 
  the terms of the GNU General Public License as published by the Free Software 
  Foundation; version 2 of the License.
  
  This program is distributed in the hope that it will be useful, but WITHOUT ANY 
  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS 
  FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
  
  You should have received a copy of the GNU General Public License along with this 
  program; if not, write to the Free Software Foundation, Inc., 51 Franklin St, Fifth 
  Floor, Boston, MA 02110-1301  USA
*/
#include "my_json_udf.h"

long long json_valid(UDF_INIT *initid, UDF_ARGS *args, char *is_null, char *error)
{
	//if args->args[0] is NULL, no need to parse
	//we can not return this at init state, because this is not an error
	if (!args->args[0])
	{
    jdump(JSON_NOTE, JSON_UDF_ERROR, "First argument is NULL");
		return false;
	}
	
	char *doc= new char[args->lengths[0] + 1];
  memcpy(doc, args->args[0], args->lengths[0]);
  doc[args->lengths[0]]= '\0';
  
	bool result= json_valid(doc);
	delete[] doc;
	if (result)
		return 1;
	else
		return 0;
}


my_bool json_valid_init(UDF_INIT *initid, UDF_ARGS *args, char *message)
{
	if (args->arg_count != 1 || args->arg_type[0] != STRING_RESULT) {
	  strcpy(message,"JSON_VALID only takes 1 string argument");
	  return 1;
	}

	initid->maybe_null = 1;
	initid->max_length = 1;
	initid->const_item = 0;
	
	return 0;
}


void json_valid_deinit(UDF_INIT *initid)
{
	return json_deinit(initid);
}


char* json_search(UDF_INIT *initid, UDF_ARGS *args, char *result, unsigned long *length, char *is_null, char *error)
{
	if (!args->args[0])
	{
    jdump(JSON_NOTE, JSON_UDF_ERROR, "First argument is NULL");
    return NULL;
  }
  
  if (!args->args[1] && 0 < args->lengths[1])
  {
    jdump(JSON_NOTE, JSON_UDF_ERROR, "Second argument is NULL, but args->lengths[1] is greater than 0, resetting");
    args->lengths[1]= 0;
  }
  
  int search_pos= 1;
  
	char *doc= new char[args->lengths[0] + 1];
  memcpy(doc, args->args[0], args->lengths[0]);
  doc[args->lengths[0]]= '\0';
  
	char* value= new char[args->lengths[1] + 1];
  memcpy(value, args->args[1], args->lengths[1]);
  value[args->lengths[1]]= '\0';
  
  if (3 == args->arg_count)
    search_pos= *((long long*) args->args[2]);
  
	char* result_string= json_search(doc, value, search_pos, true);
	jdump(JSON_NOTE, JSON_UDF_ERROR, "result string: %s", result_string);

	delete[] doc;
	delete[] value;
	
  if (result_string)
  {
    if (MAXRES < strlen(result_string))
    {
      free(initid->ptr);
      if (!(initid->ptr= (char*)malloc(strlen(result_string))))
      {
        /* malloc error, no way to return proper error */
        delete[] result_string;
        jdump(JSON_ERROR, JSON_UDF_ERROR, "Cannot allocate memory for result string");
        return NULL;
      }
    }
    memcpy(initid->ptr, result_string, strlen(result_string));
    *length= (unsigned long)strlen(result_string);
    delete[] result_string;
    
    return initid->ptr;
  }
  else
    return NULL;
}


my_bool json_search_init(UDF_INIT *initid, UDF_ARGS *args, char *message)
{
	if (args->arg_count < 2 || args->arg_count > 3)
  {
    strcpy(message,"You must supply 2 required string arguments and 1 optional integer argument");
    return 1;
  }

  for (unsigned int i= 0; i < 2; i++)
  {
    if (args->arg_type[i] != STRING_RESULT)
    {   
      strcpy(message,"First two arguments for JSON_SEARCH must be strings");
      return 1;
    }   
  }
  
  if (3 == args->arg_count && args->arg_type[2] != INT_RESULT)
  {
    strcpy(message,"Optional argument to JSON_SEARCH must be integer");
    return 1;
  }

  if (!(initid->ptr= (char*)malloc(MAXRES)))
  {
    strcpy(message,"Couldn't allocate memory for result buffer");
    return 1;
  }
  memset(initid->ptr, '\0', MAXRES);
  
  initid->maybe_null= 1;
  initid->const_item= 0;
	
	return 0;
}


void json_search_deinit(UDF_INIT *initid)
{
	return json_deinit(initid);
}


char* json_extract(UDF_INIT *initid, UDF_ARGS *args, char *result, unsigned long *length, char *is_null, char *error)
{
	if (!args->args[0])
	{
    jdump(JSON_NOTE, JSON_UDF_ERROR, "First argument is NULL");
    return NULL;
  }
  
	if (!args->args[1] || 0 == args->lengths[1])
	{
    jdump(JSON_NOTE, JSON_UDF_ERROR, "Path cannot be NULL or empty. If you want extract root element use \"$\"");
    return NULL;
  }
  
  if ('$' != args->args[1][0])
	{
    jdump(JSON_NOTE, JSON_UDF_ERROR, "Path must start from \"$\" sign");
    return NULL;
  }
  
  JKeyHolder* jeargs= make_jeargs_from_path(args->args[1], args->lengths[1]);
  
  if (!jeargs)
    return NULL;
  
	char *doc= new char[args->lengths[0] + 1];
  memcpy(doc, args->args[0], args->lengths[0]);
  doc[args->lengths[0]]= '\0';
  
	char* result_string= json_extract(doc, jeargs->array, jeargs->size);
	
  delete jeargs;
  delete[] doc;

  if (result_string)
  {
    if (MAXRES < strlen(result_string))
    {
      free(initid->ptr);
      if (!(initid->ptr= (char*)malloc(strlen(result_string))))
      {
        /* malloc error, no way to return proper error */
        delete[] result_string;
        jdump(JSON_ERROR, JSON_UDF_ERROR, "Cannot allocate memory for result string");
        return NULL;
      }
    }
    memcpy(initid->ptr, result_string, strlen(result_string));
    *length= (unsigned long)strlen(result_string);
    delete[] result_string;
    
    return initid->ptr;
  }
  else
    return NULL;
}


my_bool json_extract_init(UDF_INIT *initid, UDF_ARGS *args, char *message)
{
	if (args->arg_count != 2)
  {
    strcpy(message,"You must supply 2 arguments");
    return 1;
  }

  for (unsigned int i= 0; i < args->arg_count; i++)
  {
    if (args->arg_type[i] != STRING_RESULT 
      )
    {
      strcpy(message,"JSON_EXTRACT accepts only string arguments");
      return 1;
    }   
  }

  if (!(initid->ptr= (char*)malloc(MAXRES)))
  {
    strcpy(message,"Couldn't allocate memory for result buffer");
    return 1;
  }
  memset(initid->ptr, '\0', MAXRES);
  
  initid->maybe_null= 1;
  initid->const_item= 0;
	
	return 0;
}


void json_extract_deinit(UDF_INIT *initid)
{
	return json_deinit(initid);
}


char* json_replace(UDF_INIT *initid, UDF_ARGS *args, char *result, unsigned long *length, char *is_null, char *error)
{
	if (!args->args[0])
	{
    jdump(JSON_NOTE, JSON_UDF_ERROR, "First argument is NULL");
    return NULL;
  }
  
  //JKey** jeargs= make_jeargs(args, 1, args->arg_count - 2);
  
	if (!args->args[1] || 0 == args->lengths[1])
	{
    jdump(JSON_NOTE, JSON_UDF_ERROR, "Path cannot be NULL or empty. If you want extract root element use \"$\"");
    return NULL;
  }
  
  if ('$' != args->args[1][0])
	{
    jdump(JSON_NOTE, JSON_UDF_ERROR, "Path must start from \"$\" sign");
    return NULL;
  }
  
  JKeyHolder* jeargs= make_jeargs_from_path(args->args[1], args->lengths[1]);
  
  if (!jeargs)
    return NULL;
  
	char* doc= new char[args->lengths[0] + 1];
  memcpy(doc, args->args[0], args->lengths[0]);
  doc[args->lengths[0]]= '\0';
  
  char* new_value= create_new_value(args, args->arg_count - 1);
  
  if (!new_value)
  {
    jdump(JSON_ERROR, JSON_UDF_ERROR, "Cannot create new value for JSON_REPLACE");
    delete jeargs;
    delete[] doc;
    return NULL;
  }

	char* result_string= json_replace(doc, jeargs->array, jeargs->size, new_value);
	
	/* for (unsigned int i= 0; i < args->arg_count - 2; i++)
  {
    delete jeargs[i];
  } */

  delete jeargs;
  delete[] doc;
  delete[] new_value;

  if (result_string)
  {
    if (MAXRES < strlen(result_string))
    {
      free(initid->ptr);
      if (!(initid->ptr= (char*)malloc(strlen(result_string))))
      {
        /* malloc error, no way to return proper error */
        delete[] result_string;
        jdump(JSON_ERROR, JSON_UDF_ERROR, "Cannot allocate memory for result string");
        return NULL;
      }
    }
    memcpy(initid->ptr, result_string, strlen(result_string));
    *length= (unsigned long)strlen(result_string);
    delete[] result_string;
    
    return initid->ptr;
  }
  else
    return NULL;
}


my_bool json_replace_init(UDF_INIT *initid, UDF_ARGS *args, char *message)
{
	if (args->arg_count != 3)
  {
    strcpy(message,"You must supply 3 arguments");
    return 1;
  }

  for (unsigned int i= 0; i < args->arg_count - 1; i++)
  {
    if (args->arg_type[i] != STRING_RESULT 
      )
    {   
      strcpy(message,"JSON_REPLACE accepts only string arguments");
      return 1;
    }   
  }

  if (!(initid->ptr= (char*)malloc(MAXRES)))
  {
    strcpy(message,"Couldn't allocate memory for result buffer");
    return 1;
  }
  memset(initid->ptr, '\0', MAXRES);
  
  initid->maybe_null= 1;
  initid->const_item= 0;
	
	return 0;
}


void json_replace_deinit(UDF_INIT *initid)
{
	return json_deinit(initid);
}


char* json_append(UDF_INIT *initid, UDF_ARGS *args, char *result, unsigned long *length, char *is_null, char *error)
{
	if (!args->args[0])
	{
    jdump(JSON_NOTE, JSON_UDF_ERROR, "First argument is NULL");
    return NULL;
  }
  
	if (!args->args[1] || 0 == args->lengths[1])
	{
    jdump(JSON_NOTE, JSON_UDF_ERROR, "Path cannot be NULL or empty. If you want extract root element use \"$\"");
    return NULL;
  }
  
  if ('$' != args->args[1][0])
	{
    jdump(JSON_NOTE, JSON_UDF_ERROR, "Path must start from \"$\" sign");
    return NULL;
  }
  
  JKeyHolder* jeargs= make_jeargs_from_path(args->args[1], args->lengths[1]);
  
  if (!jeargs)
    return NULL;
  
	char* doc= new char[args->lengths[0] + 1];
  memcpy(doc, args->args[0], args->lengths[0]);
  doc[args->lengths[0]]= '\0';
  
  char* new_value= create_new_value(args, args->arg_count - 1);
  
  if (!new_value)
  {
    jdump(JSON_ERROR, JSON_UDF_ERROR, "Cannot create new value for JSON_APPEND");
    delete jeargs;
    delete[] doc;
    return NULL;
  }

	char* result_string= json_append(doc, jeargs->array, jeargs->size, new_value);
	
  delete jeargs;
  delete[] doc;
  delete[] new_value;

  if (result_string)
  {
    if (MAXRES < strlen(result_string))
    {
      free(initid->ptr);
      if (!(initid->ptr= (char*)malloc(strlen(result_string))))
      {
        /* malloc error, no way to return proper error */
        delete[] result_string;
        jdump(JSON_ERROR, JSON_UDF_ERROR, "Cannot allocate memory for result string");
        return NULL;
      }
    }
    memcpy(initid->ptr, result_string, strlen(result_string));
    *length= (unsigned long)strlen(result_string);
    delete[] result_string;
    
    return initid->ptr;
  }
  else
    return NULL;
}


my_bool json_append_init(UDF_INIT *initid, UDF_ARGS *args, char *message)
{
	if (args->arg_count != 3)
  {
    strcpy(message,"You must supply 3 arguments");
    return 1;
  }

  for (unsigned int i= 0; i < args->arg_count - 1; i++)
  {
    if (args->arg_type[i] != STRING_RESULT 
      )
    {   
      strcpy(message,"JSON_APPEND accepts only string arguments");
      return 1;
    }   
  }
  
  if (!(initid->ptr= (char*)malloc(MAXRES)))
  {
    strcpy(message,"Couldn't allocate memory for result buffer");
    return 1;
  }
  memset(initid->ptr, '\0', MAXRES);
  
  initid->maybe_null= 1;
  initid->const_item= 0;
	
	return 0;
}


void json_append_deinit(UDF_INIT *initid)
{
	return json_deinit(initid);
}


char* json_remove(UDF_INIT *initid, UDF_ARGS *args, char *result, unsigned long *length, char *is_null, char *error)
{
	if (!args->args[0])
	{
    jdump(JSON_NOTE, JSON_UDF_ERROR, "First argument is NULL");
    return NULL;
  }
  
	if (!args->args[1] || 0 == args->lengths[1])
	{
    jdump(JSON_NOTE, JSON_UDF_ERROR, "Path cannot be NULL or empty. If you want extract root element use \"$\"");
    return NULL;
  }
  
  if ('$' != args->args[1][0])
	{
    jdump(JSON_NOTE, JSON_UDF_ERROR, "Path must start from \"$\" sign");
    return NULL;
  }
  
  JKeyHolder* jeargs= make_jeargs_from_path(args->args[1], args->lengths[1]);
  
  if (!jeargs)
    return NULL;
  
	char* doc= new char[args->lengths[0] + 1];
  memcpy(doc, args->args[0], args->lengths[0]);
  doc[args->lengths[0]]= '\0';
  
	char* result_string= json_remove(doc, jeargs->array, jeargs->size);
	
  delete jeargs;
  delete[] doc;

  if (result_string)
  {
    if (MAXRES < strlen(result_string))
    {
      free(initid->ptr);
      if (!(initid->ptr= (char*)malloc(strlen(result_string))))
      {
        /* malloc error, no way to return proper error */
        delete[] result_string;
        jdump(JSON_ERROR, JSON_UDF_ERROR, "Cannot allocate memory for result string");
        return NULL;
      }
    }
    memcpy(initid->ptr, result_string, strlen(result_string));
    *length= (unsigned long)strlen(result_string);
    delete[] result_string;
    
    return initid->ptr;
  }
  else
    return NULL;
}


my_bool json_remove_init(UDF_INIT *initid, UDF_ARGS *args, char *message)
{
	if (args->arg_count != 2)
  {
    strcpy(message,"You must supply 2 arguments");
    return 1;
  }

  for (unsigned int i= 0; i < args->arg_count; i++)
  {
    if (args->arg_type[i] != STRING_RESULT 
      )
    {   
      strcpy(message,"JSON_REMOVE accepts only string arguments");
      return 1;
    }   
  }

  if (!(initid->ptr= (char*)malloc(MAXRES)))
  {
    strcpy(message,"Couldn't allocate memory for result buffer");
    return 1;
  }
  memset(initid->ptr, '\0', MAXRES);
  
  initid->maybe_null= 1;
  initid->const_item= 0;
	
	return 0;
}


void json_remove_deinit(UDF_INIT *initid)
{
	return json_deinit(initid);
}


char* json_set(UDF_INIT *initid, UDF_ARGS *args, char *result, unsigned long *length, char *is_null, char *error)
{
	if (!args->args[0])
	{
    jdump(JSON_NOTE, JSON_UDF_ERROR, "First argument is NULL");
    return NULL;
  }
  
	if (!args->args[1] || 0 == args->lengths[1])
	{
    jdump(JSON_NOTE, JSON_UDF_ERROR, "Path cannot be NULL or empty. If you want extract root element use \"$\"");
    return NULL;
  }
  
  if ('$' != args->args[1][0])
	{
    jdump(JSON_NOTE, JSON_UDF_ERROR, "Path must start from \"$\" sign");
    return NULL;
  }
  
  JKeyHolder* jeargs= make_jeargs_from_path(args->args[1], args->lengths[1]);
  
	char* doc= new char[args->lengths[0] + 1];
  memcpy(doc, args->args[0], args->lengths[0]);
  doc[args->lengths[0]]= '\0';
  
  char* new_value= create_new_value(args, args->arg_count - 1);
  
  if (!new_value)
  {
    jdump(JSON_ERROR, JSON_UDF_ERROR, "Cannot create new value for JSON_SET");
    delete jeargs;
    delete[] doc;
    return NULL;
  }

	char* result_string= json_set(doc, jeargs->array, jeargs->size, new_value);
	
  delete jeargs;
  delete[] doc;
  delete[] new_value;

  if (result_string)
  {
    if (MAXRES < strlen(result_string))
    {
      free(initid->ptr);
      if (!(initid->ptr= (char*)malloc(strlen(result_string))))
      {
        /* malloc error, no way to return proper error */
        delete[] result_string;
        jdump(JSON_ERROR, JSON_UDF_ERROR, "Cannot allocate memory for result string");
        return NULL;
      }
    }
    memcpy(initid->ptr, result_string, strlen(result_string));
    *length= (unsigned long)strlen(result_string);
    delete[] result_string;
    
    return initid->ptr;
  }
  else
    return NULL;
}


my_bool json_set_init(UDF_INIT *initid, UDF_ARGS *args, char *message)
{
	if (args->arg_count < 3)
  {
    strcpy(message,"You must supply 3 arguments");
    return 1;
  }

  for (unsigned int i= 0; i < args->arg_count - 1; i++)
  {
    if (args->arg_type[i] != STRING_RESULT 
      )
    {   
      strcpy(message,"JSON_SET accepts only string arguments");
      return 1;
    }   
  }

  if (!(initid->ptr= (char*)malloc(MAXRES)))
  {
    strcpy(message,"Couldn't allocate memory for result buffer");
    return 1;
  }
  memset(initid->ptr, '\0', MAXRES);
  
  initid->maybe_null= 1;
  initid->const_item= 0;
	
	return 0;
}


void json_set_deinit(UDF_INIT *initid)
{
	return json_deinit(initid);
}


char* json_merge(UDF_INIT *initid, UDF_ARGS *args, char *result, unsigned long *length, char *is_null, char *error)
{
	return json_merge_do(initid, args, result, length, is_null, error);
}


my_bool json_merge_init(UDF_INIT *initid, UDF_ARGS *args, char *message)
{
	return json_merge_do_init(initid, args, message);
}


void json_merge_deinit(UDF_INIT *initid)
{
	return json_deinit(initid);
}


char* json_safe_merge(UDF_INIT *initid, UDF_ARGS *args, char *result, unsigned long *length, char *is_null, char *error)
{
	return json_merge_do(initid, args, result, length, is_null, error, true);
}


my_bool json_safe_merge_init(UDF_INIT *initid, UDF_ARGS *args, char *message)
{
	return json_merge_do_init(initid, args, message, true);
}


void json_safe_merge_deinit(UDF_INIT *initid)
{
	return json_deinit(initid);
}


char* json_deep_merge(UDF_INIT *initid, UDF_ARGS *args, char *result, unsigned long *length, char *is_null, char *error)
{
	return json_merge_do(initid, args, result, length, is_null, error, false, true);
}


my_bool json_deep_merge_init(UDF_INIT *initid, UDF_ARGS *args, char *message)
{
	return json_merge_do_init(initid, args, message, true, true);
}


void json_deep_merge_deinit(UDF_INIT *initid)
{
	return json_deinit(initid);
}


long long json_contains_key(UDF_INIT *initid, UDF_ARGS *args, char *is_null, char *error)
{
  if (!args->args[0])
  {
    jdump(JSON_NOTE, JSON_UDF_ERROR, "First argument is NULL");
    return false;
  }
  
	if (!args->args[1] || 0 == args->lengths[1])
	{
    jdump(JSON_NOTE, JSON_UDF_ERROR, "Path cannot be NULL or empty. If you want extract root element use \"$\"");
    return 0;
  }
  
  if ('$' != args->args[1][0])
	{
    jdump(JSON_NOTE, JSON_UDF_ERROR, "Path must start from \"$\" sign");
    return 0;
  }
  
  JKeyHolder* jeargs= make_jeargs_from_path(args->args[1], args->lengths[1]);
  
	char *doc= new char[args->lengths[0] + 1];
  memcpy(doc, args->args[0], args->lengths[0]);
  doc[args->lengths[0]]= '\0';
  
	bool result= json_contains_key(doc, jeargs->array, jeargs->size);

  delete jeargs;
  delete[] doc;

  if (result)
    return 1;
  else
    return 0;
}


my_bool json_contains_key_init(UDF_INIT *initid, UDF_ARGS *args, char *message)
{
  if (args->arg_count != 2)
  {
    strcpy(message,"You must supply 2 arguments");
    return 1;
  }

  for (unsigned int i= 0; i < args->arg_count; i++)
  {
    if (args->arg_type[i] != STRING_RESULT 
      )
    {   
      strcpy(message,"JSON_CONTAINS_KEY accepts only string arguments");
      return 1;
    }   
  }

  initid->maybe_null= 1;
  initid->const_item= 0;
	
	return 0;
}


void json_contains_key_deinit(UDF_INIT *initid)
{
	return json_deinit(initid);	
}

char* json_test_parser(UDF_INIT *initid, UDF_ARGS *args, char *result, unsigned long *length, char *is_null, char *error)
{
  if (NULL == args->args[0] && 0 != args->lengths[0])
  {
    jdump(JSON_ERROR, JSON_UDF_ERROR, "Broken argument received");
    return NULL;
  }
  
  char *doc= new char[args->lengths[0] + 1];
  memcpy(doc, args->args[0], args->lengths[0]);
  doc[args->lengths[0]]= '\0';
  
  char* result_string= json_test_parser(doc);
  delete[] doc;
  
  if (result_string)
  {
    if (MAXRES < strlen(result_string))
    {
      free(initid->ptr);
      if (!(initid->ptr= (char*)malloc(strlen(result_string))))
      {
        /* malloc error, no way to return proper error */
        delete[] result_string;
        jdump(JSON_ERROR, JSON_UDF_ERROR, "Cannot allocate memory for result string");
        return NULL;
      }
    }
    memcpy(initid->ptr, result_string, strlen(result_string));
    *length= (unsigned long)strlen(result_string);
    delete[] result_string;
    
    return initid->ptr;
  }
  else
  {
    delete[] result_string;
    return NULL;
  }
}


my_bool json_test_parser_init(UDF_INIT *initid, UDF_ARGS *args, char *message)
{
	if (args->arg_count != 1 || args->arg_type[0] != STRING_RESULT) {
	  strcpy(message,"JSON_TEST_PARSER only takes 1 string argument\n");
	  return 1;
	}

  if (!(initid->ptr= (char*)malloc(MAXRES)))
  {
    strcpy(message,"Couldn't allocate memory for result buffer for JSON_TEST_PARSER\n");
    return 1;
  }
  memset(initid->ptr, '\0', MAXRES);

	initid->maybe_null = 1;
	initid->max_length = 1;
	initid->const_item = 0;
	
	return 0;
}


void json_test_parser_deinit(UDF_INIT *initid)
{
	return json_deinit(initid);
}


long long json_depth(UDF_INIT *initid, UDF_ARGS *args, char *is_null, char *error)
{
	//if args->args[0] is NULL, no need to parse
	//we can not return this at init state, because this is not an error
	if (!args->args[0])
	{
    jdump(JSON_NOTE, JSON_UDF_ERROR, "First argument is NULL");
		return false;
	}
	
	char *doc= new char[args->lengths[0] + 1];
  memcpy(doc, args->args[0], args->lengths[0]);
  doc[args->lengths[0]]= '\0';
  
	int result= json_depth(doc);
	delete[] doc;
	
	return result;
}


my_bool json_depth_init(UDF_INIT *initid, UDF_ARGS *args, char *message)
{
	if (args->arg_count != 1 || args->arg_type[0] != STRING_RESULT) {
	  strcpy(message,"JSON_DEPTH only takes 1 string argument");
	  return 1;
	}

	initid->maybe_null = 1;
	initid->max_length = 1;
	initid->const_item = 0;
	
	return 0;
}


void json_depth_deinit(UDF_INIT *initid)
{
	return json_deinit(initid);
}

long long json_count(UDF_INIT *initid, UDF_ARGS *args, char *is_null, char *error)
{
  JKeyHolder* jeargs;
  
	//if args->args[0] is NULL, no need to parse
	//we can not return this at init state, because this is not an error
	if (!args->args[0])
	{
    jdump(JSON_NOTE, JSON_UDF_ERROR, "First argument is NULL");
		return false;
	}
	
	if (1 < args->arg_count)
	{
	  if (!args->args[1] || 0 == args->lengths[1])
    {
      jdump(JSON_NOTE, JSON_UDF_ERROR, "Path cannot be NULL or empty. If you want extract root element use \"$\"");
      return 0;
    }
    
    if ('$' != args->args[1][0])
    {
      jdump(JSON_NOTE, JSON_UDF_ERROR, "Path must start from \"$\" sign");
      return 0;
    }
    
    jeargs= make_jeargs_from_path(args->args[1], args->lengths[1]);
	}
	else
	{
	  jeargs= new JKeyHolder(NULL, 0);
	}
	
	char *doc= new char[args->lengths[0] + 1];
  memcpy(doc, args->args[0], args->lengths[0]);
  doc[args->lengths[0]]= '\0';
  
	int result= json_count(doc, jeargs->array, jeargs->size);
	
	delete jeargs;
	delete[] doc;
	
	return result;
}


my_bool json_count_init(UDF_INIT *initid, UDF_ARGS *args, char *message)
{
	if (args->arg_count < 1) {
	  strcpy(message,"You must supply at least 1 argument");
	  return 1;
	}
	
  for (unsigned int i= 0; i < args->arg_count; i++)
  {
    if (args->arg_type[i] != STRING_RESULT 
      )
    {   
      strcpy(message,"JSON_COUNT accepts only string arguments");
      return 1;
    }   
  }

	initid->maybe_null = 1;
	initid->max_length = 1;
	initid->const_item = 0;
	
	return 0;
}


void json_count_deinit(UDF_INIT *initid)
{
	return json_deinit(initid);
}



char* json_version(UDF_INIT *initid, UDF_ARGS *args, char *result, unsigned long *length, char *is_null, char *error)
{
  char* result_string= json_version();
  
  if (result_string)
  {
    if (MAXRES < strlen(result_string))
    {
      free(initid->ptr);
      if (!(initid->ptr= (char*)malloc(strlen(result_string))))
      {
        /* malloc error, no way to return proper error */
        delete[] result_string;
        jdump(JSON_ERROR, JSON_UDF_ERROR, "Cannot allocate memory for result string");
        return NULL;
      }
    }
    memcpy(initid->ptr, result_string, strlen(result_string));
    *length= (unsigned long)strlen(result_string);
    delete[] result_string;
    
    return initid->ptr;
  }
  else
  {
    delete[] result_string;
    return NULL;
  }
}


my_bool json_version_init(UDF_INIT *initid, UDF_ARGS *args, char *message)
{
	if (args->arg_count != 0) {
	  strcpy(message,"JSON_VERSION takes no argument\n");
	  return 1;
	}

  if (!(initid->ptr= (char*)malloc(MAXRES)))
  {
    strcpy(message,"Couldn't allocate memory for result buffer for JSON_VERSION\n");
    return 1;
  }
  memset(initid->ptr, '\0', MAXRES);

	initid->maybe_null = 0;
	initid->const_item = 0;
	
	return 0;
}


void json_version_deinit(UDF_INIT *initid)
{
	return json_deinit(initid);
}


//helpers

void json_deinit(UDF_INIT *initid)
{
	if (initid->ptr)
		free(initid->ptr);
}


JKeyHolder* make_jeargs_from_path(char* path, unsigned int length)
{
  std::vector<JKey*>* result= new std::vector<JKey*>;
  std::string* keystore= new std::string("");
  std::string* arrstore= new std::string("");
  bool is_key= false;
  bool is_array= false;
  
  for (unsigned int i= 1; i < length; i++)
  {
    switch(path[i])
    {
    case '.':
      if (1 < i)
      {
        if (is_key)
        {
          if (0 < keystore->length() && ('\\' == keystore->at(keystore->length() - 1))
            && (1 == keystore->length() || ('\\' != keystore->at(keystore->length() - 2)))
            )
          {
            keystore->replace(keystore->length() - 1, 1, 1, path[i]);
          }
          else
          {
            char* res= new char[keystore->length() + 1];
            std::size_t reslen= keystore->copy(res, keystore->length());
            res[reslen]= '\0';
            result->push_back(new JKey_string(res));
            keystore->erase(0, keystore->length());
          }
        }
        else
        {
          is_key= true;
        }
      }
      else
      {
        is_key= true;
      }
      break;
    case '[':
      jdump(JSON_ERROR, JSON_UDF_ERROR, "In make_jeargs_from_path [");
      if (1 < i)
      {
        if (is_key)
        {
          if (0 < keystore->length() && ('\\' == keystore->at(keystore->length() - 1))
            && (1 == keystore->length() || ('\\' != keystore->at(keystore->length() - 2)))
            )
          {
            keystore->replace(keystore->length() - 1, 1, 1, path[i]);
          }
          else
          {
            char* res= new char[keystore->length() + 1];
            std::size_t reslen= keystore->copy(res, keystore->length());
            res[reslen]= '\0';
            result->push_back(new JKey_string(res));
            keystore->erase(0, keystore->length());
            is_key= false;
            is_array= true;
          }
        }
        else
        {
          jdump(JSON_ERROR, JSON_UDF_ERROR, "Path parse error");
          delete keystore;
          delete arrstore;
          delete result;
          return NULL;
        }
      }
      else
      {
        is_array= true;
      }
      break;
    case ']':
      if (is_array)
      {
        char* res= new char[arrstore->length() + 1];
        std::size_t reslen= arrstore->copy(res, arrstore->length());
        res[reslen]= '\0';
        result->push_back(new JKey_longlong(atoi(res)));
        delete[] res;
        arrstore->erase(0, arrstore->length());
        is_array= false;
      }
      else if (is_key && 0 < keystore->length() && ('\\' == keystore->at(keystore->length() - 1))
            && (1 == keystore->length() || ('\\' != keystore->at(keystore->length() - 2)))
            )
      {
        keystore->replace(keystore->length() - 1, 1, 1, path[i]);
      }
      else
      {
        jdump(JSON_ERROR, JSON_UDF_ERROR, "Path parse error");
        delete keystore;
        delete arrstore;
        delete result;
        return NULL;
      }
      break;
    default:
      if (1 == i)
      {
        jdump(JSON_ERROR, JSON_UDF_ERROR, "Path parse error");
        delete keystore;
        delete arrstore;
        delete result;
        return NULL;
      }
      else if (is_key)
      {
        keystore->append(1, path[i]);
      }
      else if (is_array)
      {
        switch(path[i])
        {
        case '0':
        case '1':
        case '2':
        case '3':
        case '4':
        case '5':
        case '6':
        case '7':
        case '8':
        case '9':
        case '-':
          arrstore->append(1, path[i]);
          break;
        default:
          jdump(JSON_ERROR, JSON_UDF_ERROR, "Path parse error");
          delete keystore;
          delete arrstore;
          delete result;
          return NULL;
        }
      }
      else
      {
        jdump(JSON_ERROR, JSON_UDF_ERROR, "Path parse error");
        delete keystore;
        delete arrstore;
        delete result;
        return NULL;
      }
    }
  }
  
  if (is_key)
  {
    char* res= new char[keystore->length() + 1];
    std::size_t reslen= keystore->copy(res, keystore->length());
    res[reslen]= '\0';
    result->push_back(new JKey_string(res));
    keystore->erase(0, keystore->length());
  }
  else if (is_array)
  {
    char* res= new char[arrstore->length()];
    std::size_t reslen= arrstore->copy(res, arrstore->length());
    res[reslen]= '\0';
    result->push_back(new JKey_longlong(atoi(res)));
    delete[] res;
    arrstore->erase(0, arrstore->length());
  }
  
  JKey** result_array= new JKey*[result->size()];
  unsigned int i= 0;
  for (std::vector<JKey*>::iterator it= result->begin(); it != result->end(); ++it)
  {
    result_array[i++]= (*it);
  }
  
  
  delete keystore;
  delete arrstore;
  
  int rsize= result->size();
  delete result;
  
  return new JKeyHolder(result_array, rsize);
};


char* json_merge_do(UDF_INIT *initid, UDF_ARGS *args, char *result, unsigned long *length, char *is_null, char *error, bool safe, bool deep)
{
  char* result_string;
  
	if (!args->args[0])
	{
    return NULL;
  }
  
  char** jeargs=new char*[args->arg_count];
  for (unsigned int i= 0; i < args->arg_count; i++)
  {
    if (!args->args[i] && 0 < args->lengths[i])
    {
      jdump(JSON_NOTE, JSON_UDF_ERROR, "Argument %d is NULL, but args->lengths[%d] is greater than 0, resetting", i, i);
      args->lengths[i]= 0;
    }
    jeargs[i]= new char[args->lengths[i] + 1];
    memcpy(jeargs[i], args->args[i], args->lengths[i]);
    jeargs[i][args->lengths[i]]= '\0';
  }

  result_string= json_merge(jeargs, args->arg_count, safe, deep);
	
	for (unsigned int i= 0; i < args->arg_count; i++)
  {
    delete[] jeargs[i];
  }

  delete[] jeargs;

  if (result_string)
  {
    if (MAXRES < strlen(result_string))
    {
      free(initid->ptr);
      if (!(initid->ptr= (char*)malloc(strlen(result_string))))
      {
        /* malloc error, no way to return proper error */
        delete[] result_string;
        jdump(JSON_ERROR, JSON_UDF_ERROR, "Cannot allocate memory for result string");
        return NULL;
      }
    }
    memcpy(initid->ptr, result_string, strlen(result_string));
    *length= (unsigned long)strlen(result_string);
    delete[] result_string;
    
    return initid->ptr;
  }
  else
  {
    return NULL;
  }
}


my_bool json_merge_do_init(UDF_INIT *initid, UDF_ARGS *args, char *message, bool safe, bool deep)
{
	if (args->arg_count < 2)
  {
    strcpy(message,"You must supply at least 2 string arguments");
    return 1;
  }

  for (unsigned int i= 0; i < args->arg_count; i++)
  {
    if (args->arg_type[i] != STRING_RESULT)
    {
      if (deep)
        strcpy(message,"JSON_DEEP_MERGE accepts only string arguments");
      else if (safe)
        strcpy(message,"JSON_SAFE_MERGE accepts only string arguments");
      else
        strcpy(message,"JSON_MERGE accepts only string arguments");
      return 1;
    }   
  }

  if (!(initid->ptr= (char*)malloc(MAXRES)))
  {
    free(initid->ptr);
    strcpy(message,"Couldn't allocate memory for result buffer");
    return 1;
  }
  memset(initid->ptr, '\0', MAXRES);
  
  initid->maybe_null= 1;
  initid->const_item= 0;
	
	return 0;
}

char* create_new_value(UDF_ARGS *args, int pos)
{
  char* new_value;
  if (!args->args[pos] && 0 < args->lengths[pos])
  {
    jdump(JSON_NOTE, JSON_UDF_ERROR, "In create_new_value: argument %d is NULL, but args->lengths[%d] is greater than 0, resetting", pos, pos);
    args->lengths[pos]= 0;
    args->arg_type[pos]= STRING_RESULT; /* Need to reset type as well to avoid rewriting handling of numbers */
  }
  
  switch(args->arg_type[pos])
  {
  case STRING_RESULT:
  case DECIMAL_RESULT:
    if (0 == args->lengths[pos])
    {
      if (args->args[pos])
      {
        new_value= new char[3];
        memcpy(new_value, "\"\"\0", 3);
      }
      else
      {
        new_value= new char[5];
        memcpy(new_value, "null\0", 5);
      }
    }
    else
    {
      new_value= new char[args->lengths[pos] + 1]; 
      memcpy(new_value, args->args[pos], args->lengths[pos]);
      new_value[args->lengths[pos]]= '\0';
    }
    break;
  case INT_RESULT:
    new_value= new char[MAX_BIGINT_WIDTH + 1];
    
    #ifdef __WIN__
      _set_output_format(_TWO_DIGIT_EXPONENT);
    #endif
    
    sprintf(new_value, "%lld",  *((long long*) args->args[pos]));
    break;
  case REAL_RESULT:
    new_value= new char[36 + 1];
    sprintf(new_value, "%e", *((double*) args->args[pos]));
    break;
  default:
    jdump(JSON_NOTE, JSON_UDF_ERROR, "Unknown type for new value");
    return NULL;
  }
  return new_value;
}
/*
* :tabSize=2:indentSize=2:noTabs=true:
* :folding=custom:collapseFolds=1:
*/
